#include <pi-address.h>
#include "Synch.h"
#include "AddrSynch.h"
#include "Comm.h"
#include "Database.h"
#include "Handle.h"
#include "assert.h"
#include "addr-lists.h"
#include "defs.h"

#define SLOW_SYNCH 0
#define FAST_SYNCH 1

#define dlpRecAttrDelOrArch (dlpRecAttrDeleted | dlpRecAttrArchived)

extern Tcl_Interp *g_interp;
extern char home_dir[256];
extern char username[256];

struct Addr_List *AddrList_l;
struct Addr_List *AddrList_p;

char custom[4][16]; /* 4 custom entries */
int customflags[4];

int get_addr_location (char *location)
{
  FILE *fp;
  char str[80];

  sprintf(location,"%s/%s/conf/jetaddr.dat",home_dir,username);
  fp = fopen(location,"r");
  fgets(str,80,fp);
  fclose(fp);
  str[strlen(str)-1]=0;
  sprintf(location,"%s/addrs", str);

  return 0;
}

void addr_ack(char *str, int i) {
#ifdef __DEBUG
  printf("%s (line %d)\n", str, i);
#endif
}

/* core routine for synch */
int addr_sync_lists(int synchType, int mode, char *logbuffer)
{
  struct Addr_List *addrptr_l, *addrptr_p, *aux;
  int flag;

  switch (synchType) {
  case PILOT2PC:
    addrptr_p = AddrList_p;
    while(addrptr_p) {
      aux = js_Addr_list_lookup(AddrList_l, addrptr_p->id);
      if (!aux) { /* doesn't exist locally */
        if (addrptr_p->attr & dlpRecAttrDelOrArch)
          /* nothing to do */
          ;
        else {
          /* new local record */
addr_ack("INS l",__LINE__);

          if (!(addrptr_p->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            addrptr_p = addrptr_p->next;
            continue;
          }

          addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Addr_list_insert(&AddrList_l, addrptr_p);
        }
      }
      else { /* exists locally */
        if (addrptr_p->attr & dlpRecAttrDelOrArch) {
          /* delete local record */
addr_ack("DELE l",__LINE__);
          js_Addr_list_remove(&AddrList_l, addrptr_p);
        }
        else {
          /* modify local record */
addr_ack("MOD l",__LINE__);
          addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Addr_list_modify(aux, addrptr_p);
        }
      }
      addrptr_p = addrptr_p->next;
    }
  break;
  case PC2PILOT:
    addrptr_l = AddrList_l;

    while(addrptr_l) {
      aux = js_Addr_list_lookup(AddrList_p, addrptr_l->id);
      if (!aux) { /* doesn't exist in pilot */
        if (addrptr_l->attr & dlpRecAttrDelOrArch) {
addr_ack("DEL l",__LINE__);
          aux = addrptr_l->next; /* remember position */
          /* delete local record */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_Addr_list_insert(&AddrList_p, addrptr_l);

          js_Addr_list_remove(&AddrList_l, addrptr_l);
          addrptr_l = aux;  /* restore pointer */

          continue;
        }
        else {
          if (!(addrptr_l->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            addrptr_l = addrptr_l->next;
            continue;
          }

          /* new pilot record */
printf("inserting p\n");
          addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Addr_list_insert(&AddrList_p, addrptr_l);
        }
      }
      else { /* exists in pilot */
        if (addrptr_l->attr & dlpRecAttrDelOrArch) {
          /* delete pilot record */
printf("deleting l, is in pilot\n");
          aux = addrptr_l->next; /* remember position */
          /* delete local record */
          js_Addr_list_remove(&AddrList_l, addrptr_l);
          addrptr_l = aux;  /* restore pointer */
          continue;
        }
        else {
          /* modify pilot record */
printf("modifying p\n");
          addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Addr_list_modify(aux, addrptr_l);
        }
      }
      addrptr_l = addrptr_l->next;
    }
  break;
  case BIDIREC:
    addrptr_l = AddrList_l;
    addrptr_p = AddrList_p;
    while (addrptr_p || addrptr_l) {
      flag = 0;
      if (addrptr_p) flag |= 0x10;
      if (addrptr_l) flag |= 0x01;

      if (flag==0x11) {
        if (addrptr_p->id > addrptr_l->id)
          flag = 0x01;
        else if (addrptr_p->id < addrptr_l->id)
          flag = 0x10;
        else /* = */
          flag = 0x11;
      }

      switch (flag) {
      case 0x11:
        if (addrptr_p->attr & dlpRecAttrDelOrArch) {
          if (addrptr_l->attr & dlpRecAttrDelOrArch) {
addr_ack("P:D L:D",__LINE__);
            aux = addrptr_l->next; /* remember position */
            js_Addr_list_remove(&AddrList_l, addrptr_l);
            addrptr_l = aux;  /* restore pointer */
        
            addrptr_p = addrptr_p->next;
          }
          else if (addrptr_l->attr & dlpRecAttrDirty) {
addr_ack("P:D L:Y",__LINE__);
            addrptr_p->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_Addr_list_modify(addrptr_p, addrptr_l);

            addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            addrptr_p = addrptr_p->next;
            addrptr_l = addrptr_l->next;
          }
          else {
addr_ack("P:D L:-",__LINE__);
            aux = addrptr_l->next; /* remember position */
            js_Addr_list_remove(&AddrList_l, addrptr_l);
            addrptr_l = aux;  /* restore pointer */

            addrptr_p = addrptr_p->next;
          }
        }
        else if (addrptr_p->attr & dlpRecAttrDirty) {
          if (addrptr_l->attr & dlpRecAttrDelOrArch) {
addr_ack("P:Y L:D",__LINE__);
            addrptr_l->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_Addr_list_modify(addrptr_l, addrptr_p);

            addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            addrptr_p = addrptr_p->next;
            addrptr_l = addrptr_l->next;
          }
          else if (addrptr_l->attr & dlpRecAttrDirty) {
addr_ack("P:Y L:Y",__LINE__);
            if (js_Addr_list_compare(addrptr_p, addrptr_l)) {
              addrptr_l->id = 0;
  
              addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
              addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
  
              js_Addr_list_insert(&AddrList_p, addrptr_l);
              js_Addr_list_insert(&AddrList_l, addrptr_l);
              addrptr_l->id = addrptr_p->id;
              js_Addr_list_modify(addrptr_l, addrptr_p);
              strcat(logbuffer,
              "Duplicating record since it was changed on both sides.\n");
            }
            
            addrptr_p = addrptr_p->next;
            addrptr_l = addrptr_l->next;
          }
          else {
addr_ack("P:Y L:-",__LINE__);
            addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_Addr_list_modify(addrptr_l, addrptr_p);

            addrptr_p = addrptr_p->next;
            addrptr_l = addrptr_l->next;
          }
        }
        else {
          if (addrptr_l->attr & dlpRecAttrDelOrArch) {
addr_ack("P:- L:D",__LINE__);
            addrptr_p->attr |= dlpRecAttrDelOrArch;  /* set deleted flag */

            aux = addrptr_l->next; /* remember position */
            js_Addr_list_remove(&AddrList_l, addrptr_l);
            addrptr_l = aux;  /* restore pointer */

            addrptr_p = addrptr_p->next;
          }
          else if (addrptr_l->attr & dlpRecAttrDirty) {
addr_ack("P:- L:Y",__LINE__);
            addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_Addr_list_modify(addrptr_p, addrptr_l);

            addrptr_p = addrptr_p->next;
            addrptr_l = addrptr_l->next;
          }
          else {
addr_ack("P:- L:-",__LINE__);
/* nothing to do */
            if (0)

/* uncomment this code and comment the above if, in case records should be
   duplicated when a slow synch occurs (records were modified with
   different PC's)
*/
/*            if (mode != FAST_SYNCH)*/
              if (js_Addr_list_compare(addrptr_p, addrptr_l)) {
                addrptr_l->id = 0;
  
                addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
                addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
  
                js_Addr_list_insert(&AddrList_p, addrptr_l);
                js_Addr_list_insert(&AddrList_l, addrptr_l);
                addrptr_l->id = addrptr_p->id;
                js_Addr_list_modify(addrptr_l, addrptr_p);
                strcat(logbuffer,
                "Duplicating record since a slow synch occured and data was changed.\n");
              }

            addrptr_p = addrptr_p->next;
            addrptr_l = addrptr_l->next;
          }
        }
      break;
      case 0x10:
        if (addrptr_p->attr & dlpRecAttrDelOrArch)
addr_ack("0x10 P:D",__LINE__)
          /* nothing to do */
          ;
        else if (addrptr_p->attr & dlpRecAttrDirty) {
addr_ack("0x10 P:Y",__LINE__);
          addrptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Addr_list_insert(&AddrList_l, addrptr_p);
        }
        else {
addr_ack("0x10 P:-",__LINE__);
          js_Addr_list_insert(&AddrList_l, addrptr_p);
        }

        addrptr_p = addrptr_p->next;
      break;
      case 0x01:
        if (addrptr_l->attr & dlpRecAttrDelOrArch) {
addr_ack("0x01 L:D",__LINE__);
          aux = addrptr_l->next; /* remember position */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_Addr_list_insert(&AddrList_p, addrptr_l);
          js_Addr_list_remove(&AddrList_l, addrptr_l);
          addrptr_l = aux;  /* restore pointer */
          continue;
        }
        else if (addrptr_l->attr & dlpRecAttrDirty) {
addr_ack("0x01 L:Y",__LINE__);
          addrptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Addr_list_insert(&AddrList_p, addrptr_l);
        }
        else {
addr_ack("0x01 L:-",__LINE__);
          if (mode != FAST_SYNCH)
            js_Addr_list_insert(&AddrList_p, addrptr_l);
        }

        addrptr_l = addrptr_l->next;
      break;
      } /*switch*/
    } /*while*/
  break;
  case DO_NOTHING:
  break;
  }
  return 0;
}

int write_addrs()
{
  char location[512];
  FILE *fp;
  int i;
  struct Addr_List *addrptr = AddrList_l;

  get_addr_location(location);

  if (!(fp = fopen(location,"w"))) {
    printf("Couldn't open %s for writing...\n",location);
    return -1; 
  }

/* write super-header (custom) */
  fprintf(fp,"%s\n%s\n%s\n%s\n%d\n%d\n%d\n%d\n",
               custom[0], custom[1], custom[2], custom[3],0,0,0,0);

  while (addrptr != NULL) {
    /* write record */
    fprintf(fp,"%ld %d %d %d",
            addrptr->id, addrptr->category, addrptr->attr,
            addrptr->showPhone);

    for (i=0; i<5; i++)
      fprintf(fp," %d", addrptr->phoneLabel[i]);
  
    fprintf(fp,"\n");
  
    for (i=0; i<19; i++)
      fprintf(fp," %d", addrptr->size[i]);

    fprintf(fp,"\n");
    
    for (i=0; i<19; i++)
      fprintf(fp,"%s\n", addrptr->size[i]?addrptr->entry[i]:"");

    addrptr = addrptr->next;
  }
  fclose(fp);

  return 0;
}


int read_addrs()
{
  char location[512];
  FILE *fp;
  char str[80];
  int i;
  struct Addr_List elem;

  get_addr_location(location);

  if (!(fp = fopen(location,"r"))) {
    printf("Couldn't open %s for reading...\n",location);
    return -1; 
  }

/* read super-header (custom) */
  fgets(custom[0],16,fp);   
  custom[0][strlen(custom[0])-1] = 0;
  fgets(custom[1],16,fp);
  custom[1][strlen(custom[1])-1] = 0;
  fgets(custom[2],16,fp);
  custom[2][strlen(custom[2])-1] = 0;
  fgets(custom[3],16,fp);
  custom[3][strlen(custom[3])-1] = 0;

  fgets(str,4,fp);
  customflags[0] = atoi(str);
  fgets(str,4,fp);   
  customflags[1] = atoi(str);
  fgets(str,4,fp);
  customflags[2] = atoi(str);
  fgets(str,4,fp); 
  customflags[3] = atoi(str);

  while (1) {
    fgets(str,1024,fp); /* read header */
  
    if (feof(fp)) break;
  
/* read addr header*/
    sscanf(str,"%ld %d %d %d %d %d %d %d %d\n",
           &elem.id, &elem.category,&elem.attr, &elem.showPhone,
           &elem.phoneLabel[0],
           &elem.phoneLabel[1],
           &elem.phoneLabel[2],
           &elem.phoneLabel[3],
           &elem.phoneLabel[4]);

/* read entry sizes */
    fgets(str,1024,fp); /* read header */
    sscanf(str," %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
          &elem.size[0], &elem.size[1], &elem.size[2], &elem.size[3],
          &elem.size[4], &elem.size[5], &elem.size[6], &elem.size[7],
          &elem.size[8], &elem.size[9], &elem.size[10], &elem.size[11],
          &elem.size[12], &elem.size[13], &elem.size[14], &elem.size[15],
          &elem.size[16], &elem.size[17], &elem.size[18]);

/* read entries */
    for (i=0; i<19; i++) {
      elem.entry[i] = (char *)malloc(elem.size[i]+1);

      fread(elem.entry[i],elem.size[i]+1,1,fp); /* read addr entry */
      elem.entry[i][elem.size[i]]=0;
    }
    
/* add addr to list */
    js_Addr_list_insert(&AddrList_l, &elem);

    for (i=0; i<19; i++)
      if (elem.entry[i]) free(elem.entry[i]);
  }
  fclose(fp);

  return 0;
}


void addr_save_categories(struct AddressAppInfo mai)
{
  FILE *fp;
  char location[256];
  int i;
  
  sprintf(location,"%s/%s/conf/jetaddr.cat", home_dir, username);
  fp = fopen(location,"w");
  for (i = 0; i < 16; i++) {
    fprintf(fp,"%s\n",mai.category.name[i]);
  }
  fclose(fp);
}



void js_AddrExec(i, attr, mai, m, siz, category, recid)
int i;
int attr;
struct AddressAppInfo mai;
struct Address m;
int siz;
int category;
recordid_t recid;
{
/*** Insert code here !! ***/

/*** Example: ***/
  int j;
  struct Addr_List elem;

  elem.id=recid;
  elem.attr=attr;
  elem.category=category;

  elem.showPhone = m.showPhone;
  for (j = 0; j < 5; j++)
    elem.phoneLabel[j] = m.phoneLabel[j];

  if (!(attr & dlpRecAttrDelOrArch)) {
/*  if (siz) {*/   /* if using archive uncomment this line */

    for (j = 0; j < 19; j++)
      if (m.entry[j]) {
        elem.size[j] = strlen(m.entry[j]);
        elem.entry[j] = strdup(m.entry[j]);
      } else {
        elem.size[j] = 0;
      }

#ifdef __DEBUG
    printf("\tSize: %d  Category: %s  ",
            siz, mai.category.name[category]);
    printf("Contents: \"%s %s\"\n",m.entry[1], m.entry[0]);
#endif
  } else {
    for (j = 0; j < 19; j++)
      elem.size[j] = 0;
  }
  js_Addr_list_insert(&AddrList_p, &elem);
}

/* mode=1 => fastsynch=YES*/
int js_AddrSynch(int sd, int mode, int synchType, char *path, char *logbuffer)
{
  int i, db, len, dlperr;
  unsigned char buffer[0xffff];
  char appblock[0xffff];
  struct AddressAppInfo mai;
  struct DBInfo info;
  int count, cnt=0, code;
  char tcl_string[256];
  struct Addr_List *addrptr = NULL, *aux;
  recordid_t ret_id;
  int size_list_p = 0;

  AddrList_l = AddrList_p = NULL;

#ifdef __DEBUG
  printf("js_AddrSynch [mode=%d synchtype='%c' path=%s]\n",
         mode, synchType, path);
#endif

  /* Open the database, store access handle in db */
  if ((dlperr=dlp_OpenDB(sd, 0, 0xC0, "AddressDB", &db)) < 0) {
    strcat(logbuffer, "AddressDB - Unable to open database.\n");
    return -1;
  }

  if (js_GetDBInfo(sd, "AddressDB", &info) < 0) {
    dlp_CloseDB(sd, db);
    return -1;
  }

  if ((dlperr=dlp_ReadAppBlock(sd, db, 0, (unsigned char *)appblock, 0xffff))
       < 0) {
    strcat(logbuffer, "AddressDB - Unable to read app block.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

  unpack_AddressAppInfo(&mai, (unsigned char *)appblock, 0xffff);

  addr_save_categories(mai);

  if ((dlperr=dlp_ReadOpenDBInfo(sd, db, &count)) < 0) {
    strcat(logbuffer, "AddressDB - Unable to open read DBInfo.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }


#ifdef __DEBUG
  printf("There are %d records\n", count);
  printf("Database modification date: %s\n",time_str(info.modifyDate));
  printf("      Database backup date: %s\n",time_str(info.backupDate));
#endif

  g_interp->result = (char *)Tcl_Alloc(4096);
  g_interp->freeProc = 0L;

/* read local addrs from PC */
  read_addrs();

  if (mode == SLOW_SYNCH) {
    for (i=0;;i++) {
      struct Address m;
      int attr,siz;
      int category;
      recordid_t recid;

      if (count > 0)
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",i*100/count);
      else
        sprintf(tcl_string, "progress -2 1 \"Synchronizing\"");

      code = Tcl_Eval(g_interp, tcl_string);

      len = dlp_ReadRecordByIndex(sd, db, i, buffer, &recid,
                                      &siz,&attr, &category);

      if (len != dlpErrNotFound && len < 0) {
        strcat(logbuffer, "AddressDB - Unable to read next record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

      if (len < 0) {
        break;
      }

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_Address(&m, buffer, len);
      js_AddrExec(i, attr, mai, m, siz, category, recid);
    } /* for */
  }
  else if (mode == FAST_SYNCH) {
    int last_modnum = get_mod_num();
    /* algoritmo heuristico para descodificacao do numero de
       records modificados -- estimativa razoavelmente aproximada */
    count = (int) ((info.modnum - last_modnum + 2) / 5);

    for (i=0;;i++) {
      struct Address m;
      int attr,siz,j;
      int category;
      recordid_t recid;

      if (count)
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
                i/count?100:i*100/count);
      else
        sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");

      code = Tcl_Eval(g_interp, tcl_string);

      len = dlp_ReadNextModifiedRec(sd, db, buffer, &recid, &j,
                                      &siz,&attr, &category);

      if (len != dlpErrNotFound && len < 0) {
        strcat(logbuffer, "AddressDB - Unable to read next modified record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

      if (len < 0) { /* no more records to read */
        count=cnt;

        sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");
        code = Tcl_Eval(g_interp, tcl_string);
        break;
      }

      cnt++;

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_Address(&m, buffer, len);
      js_AddrExec(i, attr, mai, m, siz, category, recid);
    } /* for */
  } /* fast / slow */


/*SYNC*/
  addr_sync_lists(synchType, mode, logbuffer);


/*WRITEBACK*/

  save_mod_num(info.modnum+count);

#ifdef __DEBUG
  printf("Writing back to pilot...\n");
#endif

  addrptr = AddrList_p;

  aux = AddrList_l;
  i = 0;
  size_list_p = js_Addr_list_count(AddrList_p);

  while (addrptr) {
    struct Address m;
    int j;

    i++;
    if (size_list_p > 0)
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / size_list_p));
    else
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / i));

    code = Tcl_Eval(g_interp, tcl_string);

    /* Skip deleted records */
    if (addrptr->attr & dlpRecAttrDelOrArch) {

      if ((dlperr=dlp_DeleteRecord(sd, db, 0, addrptr->id)) < 0) {
        strcat(logbuffer, "AddressDB - Unable to delete record.\n");
        addrptr = addrptr->next;
        continue;
      }

#ifdef __DEBUG
      printf("Deleting record %ld\n",addrptr->id);
#endif
      addrptr = addrptr->next;
      continue;
    }

    /* Skip dirty records */
    if (addrptr->attr & dlpRecAttrDirty) {
      addrptr = addrptr->next;
      continue;
    }


    m.showPhone = addrptr->showPhone;

    for (j = 0; j < 5; j++)
      m.phoneLabel[j] = addrptr->phoneLabel[j];

    for (j = 0; j < 19; j++)
      if (addrptr->size[j] && addrptr->entry[j])
        m.entry[j] = strdup(addrptr->entry[j]);
      else
        m.entry[j] = "";

    len = pack_Address(&m, buffer, 0xffff);

    if ((dlperr=dlp_WriteRecord(sd, db, addrptr->attr & dlpRecAttrSecret, addrptr->id,
      addrptr->category, buffer, len, &ret_id)) < 0) {
        strcat(logbuffer, "AddressDB - Unable to write record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }


    if (addrptr->id == 0) { /* pilot returns new record's id */
      aux->id = ret_id;
      aux=aux->next;
    }

    addrptr = addrptr->next;
  }

/* handle custom records */
  for (i = 0; i < 4; i++) {
    if (mai.labelRenamed[entryCustom1 + i]) {
      if (customflags[i]) {
        if (mode==PILOT2PC)
          strcpy(custom[i], mai.labels[entryCustom1 + i]);
        else
          strcpy(mai.labels[entryCustom1 + i], custom[i]);
      } else {
        if (mode==PILOT2PC)
          strcpy(custom[i], mai.labels[entryCustom1 + i]);
        else if (mode==BIDIREC)
          strcpy(custom[i], mai.labels[entryCustom1 + i]);
        else
          strcpy(mai.labels[entryCustom1 + i], custom[i]);
      }
    } else {
      if (customflags[i]) {
        if (mode==PILOT2PC)
          strcpy(custom[i], mai.labels[entryCustom1 + i]);
        else
          strcpy(mai.labels[entryCustom1 + i], custom[i]);
      }
    }
  }

  len = pack_AddressAppInfo(&mai, (unsigned char *)appblock, 0xffff);

  if ((dlperr=dlp_WriteAppBlock(sd, db, (unsigned char *)appblock, len))
       < 0) {
    strcat(logbuffer, "AddressDB - Unable to write app block.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

/* write local data back to PC */
  write_addrs();

  sprintf(tcl_string, "progress -2 %d \"Synchronizing\"", 200);
  code = Tcl_Eval(g_interp, tcl_string);

  if (synchType != PC2PILOT && synchType != DO_NOTHING)
    dlp_ResetSyncFlags(sd, db);
    
/* Close the database */
  dlp_CloseDB(sd, db);

  strcat(logbuffer, "OK Addr"); 
  switch (synchType) {
    case PC2PILOT:   strcat(logbuffer," Pilot<<-Desktop"); break;
    case PILOT2PC:   strcat(logbuffer," Pilot->>Desktop"); break;
    case BIDIREC:    strcat(logbuffer," Pilot<->Desktop"); break; 
  }
  strcat(logbuffer, "\n");

  js_Addr_list_free(&AddrList_l);
  js_Addr_list_free(&AddrList_p);

  return 0;
}
