#include "DatebookSynch.h"
#include "Comm.h"
#include "Database.h"
#include "Handle.h"
#include "assert.h"
#include "date-lists.h"
#include "defs.h"
#include <pi-datebook.h>

#define SLOW_SYNCH 0
#define FAST_SYNCH 1

#define dlpRecAttrDelOrArch (dlpRecAttrDeleted | dlpRecAttrArchived)

extern Tcl_Interp *g_interp;

struct Datebook_List *DatebookList_l;
struct Datebook_List *DatebookList_p;

#define ICAL_ID_MAXLEN 128
int weekstarts_p, weekstarts_l;
extern char *d_type_str[];
map_t *map = NULL;
struct tm *mytm = NULL;

time_t gmktime(struct tm *tt)
{
  time_t ret;

  setenv("TZ","",1);
  tt->tm_isdst = -1;
  ret = mktime(tt);
  setenv("TZ","",1);
  return ret;
}

static struct tm *glocaltime(time_t *tt)
{
  memset((char *)mytm, 0, sizeof(struct tm));
  memcpy((char *)mytm, localtime(tt), sizeof(struct tm));
  mytm->tm_isdst = -1;
  return mytm;
}



void date_ack(char *str, int i) {
#ifdef __DEBUG
  printf("%s (line %d)\n", str, i);
#endif
}

/* core routine for synch */
int date_sync_lists(int synchType, int mode, char *logbuffer)
{
  struct Datebook_List *dateptr_l, *dateptr_p, *aux;
  int flag;

  switch (synchType) {
  case PILOT2PC:
    dateptr_p = DatebookList_p;
    while(dateptr_p) {
      aux = js_Datebook_list_lookup(DatebookList_l, dateptr_p->id);
      if (!aux) { /* doesn't exist locally */
        if (dateptr_p->attr & dlpRecAttrDelOrArch)
          /* nothing to do */
          ;
        else {
          /* new local record */
date_ack("INS l",__LINE__);

          if (!(dateptr_p->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            dateptr_p = dateptr_p->next;
            continue;
          }

          dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Datebook_list_insert(&DatebookList_l, dateptr_p);
        }
      }
      else { /* exists locally */
        if (dateptr_p->attr & dlpRecAttrDelOrArch) {
          /* delete local record */
date_ack("DELE l",__LINE__);
          js_Datebook_list_remove(&DatebookList_l, dateptr_p);
        }
        else {
          /* modify local record */
date_ack("MOD l",__LINE__);
          dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Datebook_list_modify(aux, dateptr_p);
        }
      }
      dateptr_p = dateptr_p->next;
    }
  break;
  case PC2PILOT:
    dateptr_l = DatebookList_l;

    while(dateptr_l) {
      aux = js_Datebook_list_lookup(DatebookList_p, dateptr_l->id);
      if (!aux) { /* doesn't exist in pilot */
        if (dateptr_l->attr & dlpRecAttrDelOrArch) {
date_ack("DEL l",__LINE__);
          aux = dateptr_l->next; /* remember position */
          /* delete local record */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_Datebook_list_insert(&DatebookList_p, dateptr_l);

          js_Datebook_list_remove(&DatebookList_l, dateptr_l);
          dateptr_l = aux;  /* restore pointer */

          continue;
        }
        else {
          if (!(dateptr_l->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            dateptr_l = dateptr_l->next;
            continue;
          }

          /* new pilot record */
printf("inserting p\n");
          dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Datebook_list_insert(&DatebookList_p, dateptr_l);
        }
      }
      else { /* exists in pilot */
        if (dateptr_l->attr & dlpRecAttrDelOrArch) {
          /* delete pilot record */
printf("deleting l, is in pilot\n");
          aux = dateptr_l->next; /* remember position */
          /* delete local record */
          js_Datebook_list_remove(&DatebookList_l, dateptr_l);
          dateptr_l = aux;  /* restore pointer */
          continue;
        }
        else {
          /* modify pilot record */
printf("modifying p\n");
          dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Datebook_list_modify(aux, dateptr_l);
        }
      }
      dateptr_l = dateptr_l->next;
    }
  break;
  case BIDIREC:
    dateptr_l = DatebookList_l;
    dateptr_p = DatebookList_p;
    while (dateptr_p || dateptr_l) {
      flag = 0;
      if (dateptr_p) flag |= 0x10;
      if (dateptr_l) flag |= 0x01;

      if (flag==0x11) {
        if (dateptr_p->id > dateptr_l->id)
          flag = 0x01;
        else if (dateptr_p->id < dateptr_l->id)
          flag = 0x10;
        else /* = */
          flag = 0x11;
      }

      switch (flag) {
      case 0x11:
        if (dateptr_p->attr & dlpRecAttrDelOrArch) {
          if (dateptr_l->attr & dlpRecAttrDelOrArch) {
date_ack("P:D L:D",__LINE__);
            aux = dateptr_l->next; /* remember position */
            js_Datebook_list_remove(&DatebookList_l, dateptr_l);
            dateptr_l = aux;  /* restore pointer */
        
            dateptr_p = dateptr_p->next;
          }
          else if (dateptr_l->attr & dlpRecAttrDirty) {
date_ack("P:D L:Y",__LINE__);
            dateptr_p->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_Datebook_list_modify(dateptr_p, dateptr_l);

            dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            dateptr_p = dateptr_p->next;
            dateptr_l = dateptr_l->next;
          }
          else {
date_ack("P:D L:-",__LINE__);
            aux = dateptr_l->next; /* remember position */
            js_Datebook_list_remove(&DatebookList_l, dateptr_l);
            dateptr_l = aux;  /* restore pointer */

            dateptr_p = dateptr_p->next;
          }
        }
        else if (dateptr_p->attr & dlpRecAttrDirty) {
          if (dateptr_l->attr & dlpRecAttrDelOrArch) {
date_ack("P:Y L:D",__LINE__);
            dateptr_l->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_Datebook_list_modify(dateptr_l, dateptr_p);

            dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            dateptr_p = dateptr_p->next;
            dateptr_l = dateptr_l->next;
          }
          else if (dateptr_l->attr & dlpRecAttrDirty) {
date_ack("P:Y L:Y",__LINE__);
            if (js_Datebook_list_compare(dateptr_p, dateptr_l)) {
              dateptr_l->id = 0;
              
              while (1) {
                generate_uid(dateptr_l->i_id);
                if (map_lookup_i(map, dateptr_l->i_id)==-1) break;
              }

              dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
              dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
  
              js_Datebook_list_insert(&DatebookList_p, dateptr_l);
              js_Datebook_list_insert(&DatebookList_l, dateptr_l);

              js_Datebook_list_modify(dateptr_l, dateptr_p);
              strcat(logbuffer,
              "Duplicating record since it was changed on both sides.\n");
            }
            
            dateptr_p = dateptr_p->next;
            dateptr_l = dateptr_l->next;
          }
          else {
date_ack("P:Y L:-",__LINE__);
            dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_Datebook_list_modify(dateptr_l, dateptr_p);

            dateptr_p = dateptr_p->next;
            dateptr_l = dateptr_l->next;
          }
        }
        else {
          if (dateptr_l->attr & dlpRecAttrDelOrArch) {
date_ack("P:- L:D",__LINE__);
            dateptr_p->attr |= dlpRecAttrDelOrArch;  /* set deleted flag */

            aux = dateptr_l->next; /* remember position */
            js_Datebook_list_remove(&DatebookList_l, dateptr_l);
            dateptr_l = aux;  /* restore pointer */

            dateptr_p = dateptr_p->next;
          }
          else if (dateptr_l->attr & dlpRecAttrDirty) {
date_ack("P:- L:Y",__LINE__);
            dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_Datebook_list_modify(dateptr_p, dateptr_l);

            dateptr_p = dateptr_p->next;
            dateptr_l = dateptr_l->next;
          }
          else {
date_ack("P:- L:-",__LINE__);
/* nothing to do */
            if (0)

/* uncomment this code and comment the above if, in case records should be
   duplicated when a slow synch occurs (records were modified with
   different PC's)
*/
/*            if (mode != FAST_SYNCH)*/
              if (js_Datebook_list_compare(dateptr_p, dateptr_l)) {
                dateptr_l->id = 0;
  
                dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
                dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
  
                js_Datebook_list_insert(&DatebookList_p, dateptr_l);
                js_Datebook_list_insert(&DatebookList_l, dateptr_l);
                dateptr_l->id = dateptr_p->id;
                js_Datebook_list_modify(dateptr_l, dateptr_p);
                strcat(logbuffer,
                "Duplicating record since a slow synch occured and data was changed.\n"); 
              }

            dateptr_p = dateptr_p->next;
            dateptr_l = dateptr_l->next;
          }
        }
      break;
      case 0x10:
        if (dateptr_p->attr & dlpRecAttrDelOrArch)
date_ack("0x10 P:D",__LINE__)
          /* nothing to do */
          ;
        else if (dateptr_p->attr & dlpRecAttrDirty) {
date_ack("0x10 P:Y",__LINE__);
          dateptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Datebook_list_insert(&DatebookList_l, dateptr_p);
        }
        else {
date_ack("0x10 P:-",__LINE__);
          js_Datebook_list_insert(&DatebookList_l, dateptr_p);
        }

        dateptr_p = dateptr_p->next;
      break;
      case 0x01:
        if (dateptr_l->attr & dlpRecAttrDelOrArch) {
date_ack("0x01 L:D",__LINE__);
          aux = dateptr_l->next; /* remember position */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_Datebook_list_insert(&DatebookList_p, dateptr_l);
          js_Datebook_list_remove(&DatebookList_l, dateptr_l);
          dateptr_l = aux;  /* restore pointer */
          continue;
        }
        else if (dateptr_l->attr & dlpRecAttrDirty) {
date_ack("0x01 L:Y",__LINE__);
          dateptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Datebook_list_insert(&DatebookList_p, dateptr_l);
        }
        else {
date_ack("0x01 L:-",__LINE__);
          if (mode != FAST_SYNCH)
            js_Datebook_list_insert(&DatebookList_p, dateptr_l);
        }

        dateptr_l = dateptr_l->next;
      break;
      } /*switch*/
    } /*while*/
  break;
  case DO_NOTHING:
  break;
  }
  return 0;
}


/* saves a copy of the ical file */
int save_datebook_attributes()
{
  char locR[256], locW[256], line[256];
  FILE *fp, *fpw;

/* get data location */
  sprintf(locR, "%s/%s/conf/jetdate.dat", home_dir, username);
  fp = fopen(locR,"r"); 
  fgets(locR, 256, fp); /* read location of i.cal file */
  fclose(fp);
  locR[strlen(locR)-1]=0;  

/* begin writing the ical file's header */
  sprintf(locW,"%s/datebook.data", locR);
  strcat(locR,"/i.cal");
  fp = fopen(locR, "r");
  fpw = fopen(locW, "w");
  if (!fp || !fpw) return -1;

  while (1) {
    memset(line, 0, 256);
    fgets(line, 256, fp);
    if (feof(fp)) break;
    fputs(line, fpw);
  }

  fclose(fp);
  fclose(fpw);
  return 0;
}



/* writes all the local list data to disk */
int write_dates()
{
  char location[256];
  FILE *fp;
  struct Datebook_List *dateptr;
  struct i_cal ical;
  char fmt[1024];
  struct passwd *pw;
  int date_cnt = 0, i, found;
  time_t tt;
  char direction[10]=""; /* used in ComplexMonths */
  int repWeek;           /* used in ComplexMonths */

/* get data location */
  sprintf(location, "%s/%s/conf/jetdate.dat", home_dir, username);
  fp = fopen(location,"r"); 
  fgets(location, 256, fp); /* read location of i.cal file */
  fclose(fp);
  location[strlen(location)-1]=0;  

/* begin writing the ical file's header */
  strcat(location,"/i.cal");
  if (write_ical_header(location, weekstarts_l) < 0) return -1;

  dateptr = DatebookList_l;
  while (dateptr != NULL) {
    ical.start =   
      ical.length =
      ical.uid =  
      ical.owner =
      ical.contents =
      ical.text =
      ical.todo =
      ical.done =
      ical.remind = 
      ical.alarm =
      ical.hilite = NULL;
    ical.dates = NULL;
    ical.date_type = NULL;

/* format dateptr -> ical */

    ical.type = dateptr->event + 1; /* 0=undefined, 1=Appt or 2=Note */

    glocaltime(&dateptr->begin);
    sprintf(fmt, "%d", mytm->tm_hour * 60 + mytm->tm_min);
    ical.start = strdup(fmt);

    sprintf(fmt, "%d", (int)(dateptr->end - dateptr->begin) / 60);
    ical.length = strdup(fmt);

    ical.uid = strdup(dateptr->i_id);

    pw = getpwuid(getuid());
    ical.owner = strdup(pw->pw_name);

    if (dateptr->description)
      ical.contents = strdup(dateptr->description);
    else
      ical.contents = NULL;

    if (dateptr->note)
      ical.text = strdup(dateptr->note);
    else
      ical.text = NULL;

    if (dateptr->todo)
      ical.todo = strdup("");
    else
      ical.todo = NULL;

    if (dateptr->done)
      ical.done = strdup("");
    else
      ical.done = NULL;

    ical.remind = (char *)malloc(10);
    sprintf(ical.remind, "%d", dateptr->remind);

    switch (dateptr->hilite) {
      case 0: ical.hilite=strdup("always"); break;
      case 1: ical.hilite=strdup("never"); break;
      case 2: ical.hilite=strdup("expire"); break;
      case 3: ical.hilite=strdup("holiday"); break;
    }

    if (dateptr->alarm) {
      if (dateptr->advanceUnits == advDays)
        dateptr->advance = 0;

      ical.alarm = (char *)malloc(1024);
      ical.alarm[0] = 0;

      i = dateptr->advance;
      while (1) {
        sprintf(fmt, "%2d", i);
        strcat(ical.alarm, fmt);
        if (dateptr->advance == 0) break;
        i += dateptr->advance;
        if (i > 60) break;
        strcat(ical.alarm," ");
      }
    }

    date_cnt = 0;
    ical.dates = (char **)malloc(sizeof(char *) * (date_cnt+1));
    ical.date_type = (int *)malloc(sizeof(int) * (date_cnt+1));

/* handle repetition information */
    switch (dateptr->repeatType) {
      case repeatNone:
        ical.date_type[0] = Single;
        tt = dateptr->begin;
        glocaltime(&tt);
        sprintf(fmt, "%d/%d/%d",
                mytm->tm_mday,
                mytm->tm_mon + 1, mytm->tm_year + 1900);
        ical.dates[0] = strdup(fmt);
      break;
      case repeatDaily:
        ical.date_type[0] = Days;
        tt = dateptr->begin;
        glocaltime(&tt);
        sprintf(fmt, "%d/%d/%d",
                mytm->tm_mday,
                mytm->tm_mon + 1, mytm->tm_year + 1900);
        ical.dates[0] = strdup(fmt);
        ical.date_freq = dateptr->repeatFrequency;
      break;
      case repeatWeekly:
        found = 0;
        for (i = 0; i < 7; i++)
          if (dateptr->repeatDays[i]) {
            found++;
          }

        if (found > 1) {
          char fmt2[40]="";

          ical.date_type[0] = WeekDays;
          fmt[0] = 0;
          for (i = 0; i < 7; i++)
            if (dateptr->repeatDays[i]) {
              sprintf(fmt2, "%d ", i + 1);
              strcat(fmt,fmt2);
            }

          strcat(fmt,"Months  1 2 3 4 5 6 7 8 9 10 11 12");
          ical.dates[0] = strdup(fmt);
          ical.date_freq = dateptr->repeatFrequency;
        }
        else {
          ical.date_type[0] = Days;
          tt = dateptr->begin;
          glocaltime(&tt);
          sprintf(fmt, "%d/%d/%d",
                  mytm->tm_mday,
                  mytm->tm_mon + 1, mytm->tm_year + 1900);
          ical.dates[0] = strdup(fmt);
          ical.date_freq = 7 * dateptr->repeatFrequency;
        }
      break;
      case repeatMonthlyByDay:
        ical.date_type[0] = ComplexMonths;
        tt = dateptr->begin;
        glocaltime(&tt);
        if (dateptr->repeatDay > domLastSun ) {
          strcpy(direction, "Backward");
          repWeek = 1;
        } else {
          strcpy(direction, "Forward");
          repWeek = (dateptr->repeatDay / 7) + 1;
        }
        sprintf(fmt, "%d %d %d/%d/%d %s ByWeek %d",
                dateptr->repeatFrequency, repWeek,
                mytm->tm_mday,
                mytm->tm_mon + 1, mytm->tm_year + 1900,
                direction, (dateptr->repeatDay % 7) + 1);
        ical.dates[0] = strdup(fmt);
      break;
      case repeatMonthlyByDate:
        ical.date_type[0] = Months;
        tt = dateptr->begin;
        glocaltime(&tt);
        sprintf(fmt, "%d/%d/%d",
                mytm->tm_mday,
                mytm->tm_mon + 1, mytm->tm_year + 1900);
        ical.dates[0] = strdup(fmt);
        ical.date_freq = dateptr->repeatFrequency;
      break;
      case repeatYearly:
        ical.date_type[0] = Months;
        tt = dateptr->begin;
        glocaltime(&tt);
        sprintf(fmt, "%d/%d/%d",
                mytm->tm_mday,
                mytm->tm_mon + 1, mytm->tm_year + 1900);
        ical.dates[0] = strdup(fmt);
        ical.date_freq = 12 * dateptr->repeatFrequency;
      break;
    }



/* in case of a repetition, a Start line is always necessary */
    if (dateptr->repeatType != repeatNone) {
      date_cnt++;
      ical.dates = (char **)realloc(ical.dates,
                        sizeof(char *) * (date_cnt+1));
      ical.date_type = (int *)realloc(ical.date_type,
                        sizeof(int) * (date_cnt+1));
      tt = dateptr->begin;
      glocaltime(&tt);
      sprintf(fmt, "%d/%d/%d",
              mytm->tm_mday,
              mytm->tm_mon + 1, mytm->tm_year + 1900);
      ical.date_type[date_cnt] = Start;
      ical.dates[date_cnt] = strdup(fmt);
   }



/* check if Finish line is needed */
    if (!dateptr->repeatForever) {
      /* set a finish date */
      date_cnt++;
      ical.dates = (char **)realloc(ical.dates,
                        sizeof(char *) * (date_cnt+1));
      ical.date_type = (int *)realloc(ical.date_type,
                        sizeof(int) * (date_cnt+1));
      glocaltime(&dateptr->repeatEnd);
      sprintf(fmt, "%d/%d/%d",
              mytm->tm_mday,
              mytm->tm_mon + 1, mytm->tm_year + 1900);
      ical.date_type[date_cnt] = Finish;
      ical.dates[date_cnt] = strdup(fmt);
    }
    
/* check if Deleted lines are needed */

    for (i = 0; i < dateptr->exceptions; i++) {
      date_cnt++;
      ical.dates = (char **)realloc(ical.dates,
                        sizeof(char *) * (date_cnt+1));
      ical.date_type = (int *)realloc(ical.date_type,
                        sizeof(int) * (date_cnt+1));
      glocaltime(&dateptr->exception[i]);
      sprintf(fmt, "%d/%d/%d",
              mytm->tm_mday,
              mytm->tm_mon + 1, mytm->tm_year + 1900);

      ical.date_type[date_cnt] = Deleted;
      ical.dates[date_cnt] = strdup(fmt);
    }

    date_cnt++;
    ical.dates = (char **)realloc(ical.dates,
                        sizeof(char *) * (date_cnt+1));
    ical.dates[date_cnt] = NULL;

/* ical entry is ready to be written */
    write_ical_entry(location, &ical);

/* free all allocated data for this entry */
    if (ical.start) free(ical.start);
    if (ical.length) free(ical.length);
    if (ical.uid) free(ical.uid);
    if (ical.owner) free(ical.owner);
    if (ical.contents) free(ical.contents);
    if (ical.text) free(ical.text);
    if (ical.todo) free(ical.todo);
    if (ical.done) free(ical.done);
    if (ical.remind) free(ical.remind);
    if (ical.alarm) free(ical.alarm);
    if (ical.hilite) free(ical.hilite);
    if (ical.date_type) free(ical.date_type);
    if (ical.dates) {
      int k=0;
      while (ical.dates[k])
        free(ical.dates[k++]);
      free(ical.dates);
    }

/* move to next entry */
    dateptr = dateptr->next;
  }
  fclose(fp);

/* save a copy so that we can check attributes next time */
  save_datebook_attributes(location);

  return 0;
}



/* interprets the ical data in order to set appropriate attributes */
/* Note: this only handles the dirty flag in the attributes. */
int read_datebook_attributes(struct i_cal *ical)
{
  char location[256];
  struct i_cal si; /* will contain ical data after last synch */
  FILE *fp;
  int i;

  sprintf(location, "%s/%s/conf/jetdate.dat", home_dir, username);
  fp = fopen(location,"r");
  fgets(location, 256, fp); /* read location of datebook data */
  fclose(fp);
  location[strlen(location)-1]=0;
    
  strcat(location,"/datebook.data");
  fp = fopen(location,"r");
  if (!fp || feof(fp)) return dlpRecAttrDirty; /* file is empty */

  while (1) {
    memset((char *)&si, 0, sizeof(struct i_cal));
    if (read_ical_entry(fp, &si) < 0) break; /* entry was not found*/

/* Soft error detection */
    if (si.uid == NULL || ical->uid == NULL) {
#ifdef __DEBUG
      printf("Error while reading ICAL data.\n");
#endif
      fclose(fp);
      return 0;
    }

/* Lookup the correct ical entry in datebook.data file */

/* not looking for this uid, skip */
    if (strcmp(si.uid, ical->uid)) continue;

    if (si.type != ical->type) { fclose(fp); return dlpRecAttrDirty; }

    if (strcmp(si.start?si.start:"",
               ical->start?ical->start:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.length?si.length:"",
               ical->length?ical->length:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.owner?si.owner:"",
               ical->owner?ical->owner:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.contents?si.contents:"",
               ical->contents?ical->contents:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.text?si.text:"",
               ical->text?ical->text:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.todo?si.todo:"",
               ical->todo?ical->todo:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.done?si.done:"",
               ical->done?ical->done:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.remind?si.remind:"",
               ical->remind?ical->remind:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.alarm?si.alarm:"",
               ical->alarm?ical->alarm:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }
    if (strcmp(si.hilite?si.hilite:"",
               ical->hilite?ical->hilite:"")) {
      fclose(fp); return dlpRecAttrDirty;
    }

    if (si.date_freq != ical->date_freq) {
      fclose(fp); return dlpRecAttrDirty;
    }

    i = 0;
    while (si.dates[i]) {
      if ((!ical->dates[i]) || (strcmp(ical->dates[i], si.dates[i]))) {
        fclose(fp); return dlpRecAttrDirty;
      }
      if (si.date_type[i] != ical->date_type[i]) {
        fclose(fp); return dlpRecAttrDirty;
      }
      i++;
    }

/* If we get here, entry was not changed so return void attribute.
   Otherwise, changes will have been triggered above and returned. */
    fclose(fp);
    return 0; /* item was not changed */
  }


/* If we get here, entry was not found which means this is a new item */
  fclose(fp);
  return dlpRecAttrDirty; /* new item */
}


/* reads all the dta from the user's ICAL data file and inserts the data into
   the local list */
int read_dates()
{
  char location[256], line[256];
  FILE *fp;
  int count=0,i;
  struct Datebook_List elem;
  struct i_cal ical;
  int direction;

  read_ical_header(&weekstarts_l);

/* get location of i.cal file */
  sprintf(location, "%s/%s/conf/jetdate.dat", home_dir, username);
  fp = fopen(location,"r");
  fgets(location, 256, fp); /* read location of i.cal file */
  fclose(fp);
  location[strlen(location)-1]=0;
    
  strcat(location,"/i.cal");
  fp = fopen(location,"r");
  if (!fp || feof(fp)) return 0; /* file is empty */

  fgets(line,MAXLINE,fp); /* Discard first line ("Calendar [vX.Y]") */

  while (1) {
    memset((char *)&ical, 0, sizeof(struct i_cal));
    if (read_ical_entry(fp, &ical) < 0) break;

    elem.event = ical.type - 1; /* 0=undefined, 1=Appt or 2=Note */

    elem.attr = 0;
    if (ical.uid) {
      elem.i_id = strdup(ical.uid);
      elem.id = map_lookup_i(map, elem.i_id);
      if (elem.id == -1) { /* map did not contain entry -- new item */
        elem.id = 0;
        elem.attr |= dlpRecAttrDirty;
      }
      else {
        map_set_del(&map, elem.i_id, 0);
      }
    }
    else {
#ifdef __DEBUG
      printf("Strange... ical entry does not have a UID!\n");
      elem.id = -1;
#endif
    }

    if (ical.contents)
      elem.description = strdup(ical.contents);
    else {
      elem.description = (char *)malloc(1);
      elem.description[0] = 0;
    }

    if (ical.text)
      elem.note = strdup(ical.text);
    else {
      elem.note = (char *)malloc(1);
      elem.note[0] = 0;
    }

    elem.todo = ical.todo != NULL;
    elem.done = ical.done != NULL;
    if (ical.remind)
      elem.remind = atoi(ical.remind);
    else
      elem.remind = 0;

    if (ical.hilite) {
      if (!strcasecmp(ical.hilite, "always"))
        elem.hilite = 0;
      else if (!strcasecmp(ical.hilite, "never"))
        elem.hilite = 1;
      else if (!strcasecmp(ical.hilite, "expire"))
        elem.hilite = 2;
      else if (!strcasecmp(ical.hilite, "holiday"))
        elem.hilite = 3;
    }
    else
      elem.hilite = -1;

    elem.alarm = 0;
    if (ical.alarm) {
      int int1, ptr=0;
      char *fmt;

      fmt = ical.alarm;
      while (fmt[ptr] == ' ') ptr++;
      int1 = atoi(strtok(&fmt[ptr], " "));

      elem.alarm = 1;
      elem.advanceUnits = advMinutes;
      elem.advance = int1;
    }

    if (elem.remind) {
      elem.alarm = 1;
      elem.advanceUnits = advDays;
      elem.advance = elem.remind;
    }

/* default values */
    elem.repeatForever = 1;
    elem.repeatType = repeatNone;

    if (ical.dates && ical.dates[0]) {
      if (ical.date_type[0] == Empty) {
        elem.begin = elem.end = -1;
        elem.repeatType = repeatNone;
      }
      if (ical.date_type[0] == Single || ical.type == 2) {
        char *str = strdup(ical.dates[0]);
        mytm->tm_mday = atoi(strtok(str,"/"));
        mytm->tm_mon = atoi(strtok(NULL,"/")) - 1;
        mytm->tm_year = atoi(strtok(NULL,"/")) - 1900;
        mytm->tm_sec = mytm->tm_min = mytm->tm_hour = 0;
        elem.begin = gmktime(mytm);
        if (ical.type != 2) {
          elem.begin += 60 * atoi(ical.start);
          elem.end = elem.begin + 60 * atoi(ical.length);
        }
        else
          elem.end = elem.begin;

        elem.repeatType = repeatNone;
      }
      if (ical.date_type[0] == Days) {
        elem.repeatType = repeatDaily;
        elem.repeatForever = 1;
        elem.repeatFrequency = ical.date_freq;
        if (ical.date_freq % 7 == 0) {
          int j;
          struct tm *mytm2;
          char *str = strdup(ical.dates[1]);

          elem.repeatFrequency /= 7;
          elem.repeatType = repeatWeekly;
          for (j = 0; j < 7; j++) {
            elem.repeatDays[j] = 0;
          }

          mytm->tm_mday = atoi(strtok(str,"/"));
          mytm->tm_mon = atoi(strtok(NULL,"/")) - 1;
          mytm->tm_year = atoi(strtok(NULL,"/")) - 1900;
          mytm->tm_sec = mytm->tm_min = mytm->tm_hour = 0;
          elem.begin = gmktime(mytm);
          mytm2 = glocaltime(&elem.begin);
          j = mytm2->tm_wday;
          elem.repeatDays[j] = 1;
        }
      }
      if (ical.date_type[0] == Months) {
        elem.repeatType = repeatMonthlyByDate;
        elem.repeatForever = 1;
        elem.repeatFrequency = ical.date_freq;
      }
      if (ical.date_type[0] == WeekDays) {
        char *ptr;

        for (i = 0; i < 7; i++)
          elem.repeatDays[i] = 0;

        elem.repeatType = repeatWeekly;
        elem.repeatFrequency = 1;
        elem.repeatForever = 1;

        ptr = ical.dates[0];
        while (!isalpha(*ptr) && *ptr) {
          if (isdigit(*ptr)) {
            i = (int) (*ptr - '0');
            elem.repeatDays[i - 1] = 1;
          }
          ptr++;
        }
      }
      if (ical.date_type[0] == MonthDays) {
        elem.repeatType = repeatNone;
        elem.repeatForever = 0;
      }
      if (ical.date_type[0] == ComplexMonths) {
        char *ptr, *ptr2;
        char *str = strdup(ical.dates[1]);
        int week=0, dowk=0, willpunt = 0; /* week, dayOfWeek and error */

        mytm->tm_mday = atoi(strtok(str,"/"));
        mytm->tm_mon = atoi(strtok(NULL,"/")) - 1;
        mytm->tm_year = atoi(strtok(NULL,"/")) - 1900;
        mytm->tm_sec = mytm->tm_min = mytm->tm_hour = 0;
        elem.begin = gmktime(mytm);

        elem.repeatType = repeatMonthlyByDay;
        elem.repeatForever = 1;

        ptr = ical.dates[0];
        /* get frequency */
        while (ptr) {
          if (isdigit(*ptr)) {
            elem.repeatFrequency = *ptr - '0';
            ptr++;
            break;
          }
          ptr++;
        }
        /* get week */
        week = 0;
        while (*ptr == ' ') ptr++;
        while (ptr) {
          if (isdigit(*ptr))
            week = week * 10 + (*ptr - '0');

          ptr++;
          if (*ptr == ' ') { week--; break; }
        }
        while (*ptr == ' ') ptr++;
/* skip date */
        while (*ptr != ' ') ptr++;

        while (*ptr == ' ') ptr++;

        direction = 1;
        ptr2 = strstr(ptr,"Backward");
        if (ptr2 != NULL) {
          direction = -1;
          if (week == 0)
            week = 4; /* first week from end is last week */
        }
        else if ((ptr2 = strstr(ptr, "Forward")) == NULL)
          willpunt = 1; /* error */

        if (willpunt) {
          elem.repeatFrequency = 0;
          elem.repeatWeekstart = 0;
          elem.repeatForever = 0;
          elem.repeatEnd = (time_t) 0;
          elem.repeatType = repeatNone;

#ifdef __DEBUG
          printf("I don't understand the repetition for \"%s\"\n",
                elem.description);
          printf("Entering without repetition.\n");
#endif

          return -1;
        }

        ptr2 = strstr(ptr, "ByWeek");
        if (ptr2) {
          ptr = ptr2;
          ptr += strlen("ByWeek");

          /* get day of week */
          while (ptr) {
            if (isdigit(*ptr)) {
              dowk = (*ptr - '0') - 1;
              ptr++;
              break;
            }
            ptr++;
          }

          if (direction == 1)
            elem.repeatDay = (enum DayOfMonthType) 7 * week + dowk;
          else {
            int wday=0;

            switch (dowk) {
              case 0: wday = domLastSun; break;
              case 1: wday = domLastMon; break;
              case 2: wday = domLastTue; break;
              case 3: wday = domLastWen; break;
              case 4: wday = domLastThu; break;
              case 5: wday = domLastFri; break;
              case 6: wday = domLastSat; break;
            }
            week++;
            elem.repeatDay = wday - (7 * week);
          }
        }

        if ((strstr(ptr, "ByDay")) ||
           (strstr(ptr, "ByWorkDay"))) {
          elem.repeatFrequency = 0;
          elem.repeatWeekstart = 0;
          elem.repeatForever = 0;
          elem.repeatEnd = (time_t) 0;
          elem.repeatType = repeatNone;

#ifdef __DEBUG
          printf("I don't understand the repetition for \"%s\"\n",
                elem.description);
          printf("Entering without repetition.\n");
#endif

          return -1;
        }

      }
    }

    i = 1;
    elem.exceptions = 0;
    elem.exception = NULL;
    while (ical.dates[i]) {
      if (ical.date_type[i] == Start && ical.type != 2) {
        char *str = strdup(ical.dates[i]);
        mytm->tm_mday = atoi(strtok(str,"/"));
        mytm->tm_mon = atoi(strtok(NULL,"/")) - 1;
        mytm->tm_year = atoi(strtok(NULL,"/")) - 1900;
        mytm->tm_sec = mytm->tm_min = mytm->tm_hour = 0;
        elem.begin = gmktime(mytm);
        elem.begin += 60 * atoi(ical.start);
        elem.end = elem.begin + 60 * atoi(ical.length);
      }
      if (ical.date_type[i] == Finish) {
        char *str = strdup(ical.dates[i]);
        mytm->tm_mday = atoi(strtok(str,"/"));
        mytm->tm_mon = atoi(strtok(NULL,"/")) - 1;
        mytm->tm_year = atoi(strtok(NULL,"/")) - 1900;
        mytm->tm_sec = mytm->tm_min = mytm->tm_hour = 0;
        elem.repeatEnd = gmktime(mytm);
        elem.repeatForever = 0;
      }
      if (ical.date_type[i] == Deleted) {
        char *str = strdup(ical.dates[i]);
        mytm->tm_mday = atoi(strtok(str,"/"));
        mytm->tm_mon = atoi(strtok(NULL,"/")) - 1;
        mytm->tm_year = atoi(strtok(NULL,"/")) - 1900;
        mytm->tm_sec = mytm->tm_min = mytm->tm_hour = 0;

        elem.exception = (time_t *) realloc(elem.exception,
                         sizeof(time_t) * (elem.exceptions + 1));
        elem.exception[elem.exceptions] = gmktime(mytm);
        elem.exceptions++;
      }
      i++;
    }

/* set the attributes for this entry */
    elem.attr |= read_datebook_attributes(&ical);

/* add date to list */
    js_Datebook_list_insert(&DatebookList_l, &elem);

    free(elem.i_id);
    free(elem.description);
    free(elem.note);

    count++;
  }

  fclose(fp);

  return 0;
}



void insert_ical_deleted_items()
{
  map_t *ptr = map;
  struct Datebook_List elem;

  while (ptr) {
    if (ptr->deleted) {
      elem.id = ptr->p_id;
      elem.i_id = strdup(ptr->i_id);
      elem.description = NULL;
      elem.note = NULL;
      elem.attr = dlpRecAttrDeleted;
      elem.exceptions = 0;
      elem.exception = NULL;

      js_Datebook_list_insert(&DatebookList_l, &elem);
    }
    ptr = ptr->next;
  }
}


void js_DatebookExec(i, attr, mai, m, siz, category, recid)
int i;
int attr;
struct AppointmentAppInfo mai;
struct Appointment m;
int siz;
int category;
recordid_t recid;
{
/*** Insert code here !! ***/

/*** Example: ***/
#ifdef __DEBUG
  char *ptr;
#endif
  struct Datebook_List elem;

  memset((char *)&elem, 0, sizeof(struct Datebook_List));
  elem.id=recid;
  elem.attr=attr;
  elem.category=category;

  elem.i_id = (char *)malloc(ICAL_ID_MAXLEN);
  strcpy(elem.i_id, map_lookup_p(map, elem.id));
  if (!elem.i_id[0]) {
    while (1) {
      generate_uid(elem.i_id);
      if (map_lookup_i(map, elem.i_id)==-1) break;
    }
  }

  if (!(attr & dlpRecAttrDelOrArch)) {
    elem.event = m.event;

    elem.begin = gmktime(&m.begin);
    elem.end = gmktime(&m.end);

    elem.alarm = m.alarm;
    elem.advance = m.advance;
    elem.advanceUnits = m.advanceUnits;

    elem.remind = 0;
    if (m.alarm) {
      if (m.advanceUnits == advDays) {
        elem.remind = elem.advance;
        elem.advance = 0;
      }
      else if (m.advanceUnits == advHours)
        elem.advance *= 60;
    }

    elem.repeatType = m.repeatType;
    elem.repeatForever = m.repeatForever;
    elem.repeatEnd = gmktime(&m.repeatEnd);
    elem.repeatFrequency = m.repeatFrequency;
    elem.repeatDay = m.repeatDay;
    for (i = 0; i < 7; i++)
      elem.repeatDays[i] = m.repeatDays[i];
    elem.repeatWeekstart = m.repeatWeekstart;

    elem.todo = 0;
    elem.done = 0;

    elem.exceptions = m.exceptions;
    if (m.exceptions > 0)
      elem.exception = (time_t *)malloc(sizeof(time_t) * m.exceptions);
    else if (m.exceptions == 0)
      elem.exception = NULL;
    else ASSERT(0)

    for (i = 0; i < m.exceptions; i++)
      elem.exception[i] = gmktime(&m.exception[i]);

    if (m.description)
      elem.description = strdup(m.description);
    else
      elem.description = 0;

    if (m.note)
      elem.note = strdup(m.note);
    else
      elem.note = 0;

#ifdef __DEBUG
    printf("\tSize: %d  Category: %s  ",
            siz, mai.category.name[category]);
    ptr=m.description;
    while (*ptr && *ptr != '\n') ptr++;
    *ptr=0;
    printf("Contents: \"%s[...]\"\n",m.description);
#endif
  } else {
    elem.description = elem.note = 0;
    elem.exceptions = 0;
    elem.exception = NULL;
  }

  js_Datebook_list_insert(&DatebookList_p, &elem);
}


/* rebuilds the map based on the post-synch information */
void reconstruct_map(int synchType)
{
  struct Datebook_List *ptr = NULL;

/* build a new fresh map */
  map_free(&map);

  switch (synchType) {
    case BIDIREC:
    case PILOT2PC:
      ptr = DatebookList_l;
    break;
    case PC2PILOT:
      ptr = DatebookList_p;
    break;
  }

  while (ptr) {
    map_insert(&map, ptr->i_id, ptr->id);
    ptr = ptr->next;
  }
}




/* mode=1 => fastsynch=YES*/
int js_DatebookSynch(int sd, int mode, int synchType, char *path, char *logbuffer)
{
  int i, db, len, dlperr;
  unsigned char buffer[0xffff];
  char appblock[0xffff];
  struct AppointmentAppInfo mai;
  struct DBInfo info;
  int count, code;
  char tcl_string[256];
  struct Datebook_List *dateptr = NULL, *aux;
  recordid_t ret_id;
  int size_list_p = 0;
  DatebookList_l = DatebookList_p = NULL;

  mytm = (struct tm *)malloc(sizeof (struct tm));

#ifdef __DEBUG
  printf("js_DatebookSynch [mode=%d synchtype='%c' path=%s]\n",
         mode, synchType, path);
#endif

  /* Open the database, store access handle in db */
  if ((dlperr=dlp_OpenDB(sd, 0, 0xC0, "DatebookDB", &db)) < 0) {
    strcat(logbuffer, "DatebookDB - Unable to open database.\n");
    return -1;
  }

  if (js_GetDBInfo(sd, "DatebookDB", &info) < 0) {
    dlp_CloseDB(sd, db);
    return -1;
  }

  if ((dlperr=dlp_ReadAppBlock(sd, db, 0, (unsigned char *)appblock, 0xffff))
       < 0) {
    strcat(logbuffer, "DatebookDB - Unable to read app block.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

  unpack_AppointmentAppInfo(&mai, (unsigned char *)appblock, 0xffff);

  weekstarts_p = mai.startOfWeek;

  if ((dlperr=dlp_ReadOpenDBInfo(sd, db, &count)) < 0) {
    strcat(logbuffer, "DatebookDB - Unable to open read DBInfo.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }


/* read local dates from PC */
  map_read(&map); /* read ICAL-Pilot ID map */


  read_dates();
  insert_ical_deleted_items();




#ifdef __DEBUG
  printf("There are %d records\n", count);
printf("Database modification date: %s\n",time_str(info.modifyDate));
printf("      Database backup date: %s\n",time_str(info.backupDate));
#endif

  g_interp->result = (char *)Tcl_Alloc(4096);
  g_interp->freeProc = 0L;

  for (i=0;;i++) {
    struct Appointment m;
    int attr,siz;
    int category;
    recordid_t recid;

    if (count)
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",i*100/count);
    else
      sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");
    code = Tcl_Eval(g_interp, tcl_string);

    len = dlp_ReadRecordByIndex(sd, db, i, buffer, &recid,
                                      &siz, &attr, &category);

    if (len != dlpErrNotFound && len < 0) {
      strcat(logbuffer, "DatebookDB - Unable to read next record.\n");
      dlp_CloseDB(sd, db);
      return -1;
    }

    if (len < 0) {
      break;
    }

#ifdef __DEBUG
    printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

    unpack_Appointment(&m, buffer, len);
    js_DatebookExec(i, attr, mai, m, siz, category, recid);
  } /* for */


/*SYNC*/
  date_sync_lists(synchType, mode, logbuffer);

/*WRITEBACK*/

#ifdef __DEBUG
  printf("Writing back to pilot...\n");
#endif

  dateptr = DatebookList_p;

  i = 0;
  size_list_p = js_Datebook_list_count(DatebookList_p);

  while (dateptr) {
    struct Appointment m;
    int j;

    i++;
    if (size_list_p > 0)
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / size_list_p));
    else
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / i));

    code = Tcl_Eval(g_interp, tcl_string);

    /* Skip deleted records */
    if (dateptr->attr & dlpRecAttrDelOrArch) {
      if ((dlperr=dlp_DeleteRecord(sd, db, 0, dateptr->id)) < 0) {
        strcat(logbuffer, "DatebookDB - Unable to delete record.\n");
        dateptr = dateptr->next;
        continue;
      }

#ifdef __DEBUG
      printf("Deleted record %ld\n",dateptr->id);
#endif
      dateptr = dateptr->next;
      continue;
    }

    /* Skip dirty records */
    if (dateptr->attr & dlpRecAttrDirty) {
      dateptr = dateptr->next;
      continue;
    }


    m.event = dateptr->event;
    memcpy(&m.begin, glocaltime(&dateptr->begin), sizeof(struct tm));
    memcpy(&m.end, glocaltime(&dateptr->end), sizeof(struct tm));
    m.alarm = dateptr->alarm;
    m.advance = dateptr->advance;
    m.advanceUnits = dateptr->advanceUnits;

/* virtual alarm time -- reset to valid data (minutes to hours) */
    if (m.alarm && m.advanceUnits == advHours)
      m.advance /= 60;

/* pilot does not accept more than two digits for advance */
    if (m.alarm && m.advanceUnits == advMinutes && m.advance > 99) {
      m.advanceUnits = advHours;
      m.advance /= 60;
    }

    m.repeatType = dateptr->repeatType;
    m.repeatForever = dateptr->repeatForever;
    memcpy(&m.repeatEnd, glocaltime(&dateptr->repeatEnd),
           sizeof(struct tm));

    m.repeatFrequency = dateptr->repeatFrequency;
    m.repeatDay = dateptr->repeatDay;
    for (j = 0; j < 7; j++)
      m.repeatDays[j] = dateptr->repeatDays[j];

    m.repeatWeekstart = dateptr->repeatWeekstart;
    m.exceptions = dateptr->exceptions;
    if (m.exceptions > 0)
      m.exception = (struct tm *) malloc(sizeof(struct tm) * m.exceptions);
    else if (m.exceptions == 0)
      m.exception = NULL;
    else
      ASSERT(0)

    for (j = 0; j < m.exceptions; j++) {
      memcpy(&m.exception[j], glocaltime(&dateptr->exception[j]),
             sizeof(struct tm));
    }
    
    m.description = strdup(dateptr->description);
    m.note = strdup(dateptr->note);

    if (!m.note[0]) {
      free(m.note);
      m.note = NULL;
    }

    if (dateptr->remind) {
      m.advanceUnits = advDays;
      m.advance = dateptr->remind;
      m.alarm = 1;
    }

    len = pack_Appointment(&m, buffer, 0xffff);

    if ((dlperr=dlp_WriteRecord(sd, db, dateptr->attr & dlpRecAttrSecret, dateptr->id,
      dateptr->category, buffer, len, &ret_id)) < 0) {
        strcat(logbuffer, "DatebookDB - Unable to write record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

    if (m.note) free(m.note);
    free(m.description);

    if (dateptr->id == 0) { /* pilot returns new record's id */
      aux = DatebookList_l;
      while (aux) {
        if (!strcmp(aux->i_id,dateptr->i_id)) {
          aux->id = ret_id;
          break;
        }
        aux = aux->next;
      }

      dateptr->id = ret_id;
    }

    dateptr = dateptr->next;
  }

  reconstruct_map(synchType);

/* write local data back to PC */
  write_dates();

  map_write(map); /* write ICAL-Pilot ID map */

  sprintf(tcl_string, "progress -2 %d \"Synchronizing\"", 200);
  code = Tcl_Eval(g_interp, tcl_string);

  if (synchType != PC2PILOT && synchType != DO_NOTHING)
    dlp_ResetSyncFlags(sd, db);
    
/* Close the database */
  if ((code=dlp_CloseDB(sd, db)) < 0) {
#ifdef __DEBUG
    printf("dlp_CloseDB returned an error: %d (line %d)\n",code,__LINE__);
#endif
  }

  strcat(logbuffer, "OK Datebook");
  switch (synchType) {
    case PC2PILOT:   strcat(logbuffer," Pilot<<-Desktop"); break;
    case PILOT2PC:   strcat(logbuffer," Pilot->>Desktop"); break;
    case BIDIREC:    strcat(logbuffer," Pilot<->Desktop"); break;
  }
  strcat(logbuffer, "\n");
  js_Datebook_list_free(&DatebookList_l);
  js_Datebook_list_free(&DatebookList_p);

  return 0;
}
