#include "Synch.h"
#include "MailSynch.h"
#include "Comm.h"
#include "Database.h"
#include "Handle.h"
#include "assert.h"
#include "mail-lists.h"
#include "defs.h"

#define SLOW_SYNCH 0
#define FAST_SYNCH 1

#define SET_INBOX 0
#define SET_OUTBOX 1
#define SET_DELETED 2
#define SET_FILED 3

#define dlpRecAttrDelOrArch (dlpRecAttrDeleted | dlpRecAttrArchived)

extern Tcl_Interp *g_interp;
extern char home_dir[256];
extern char username[256];

struct Mail_List *MailList_l;
struct Mail_List *MailList_p;

char pop_host[256];
char pop_user[20];
char pop_pass[20];



void mail_ack(char *str, int i) {
#ifdef __DEBUG
  printf("%s (line %d)\n", str, i);
#endif
}


/* core routine for synch */
int mail_sync_lists(int synchType, int mode)
{
  struct Mail_List *mailptr_l, *mailptr_p, *aux;
  int flag;

  switch (synchType) {
  case PILOT2PC:
    mailptr_p = MailList_p;
    while(mailptr_p) {
      aux = js_Mail_list_lookup(MailList_l, mailptr_p->date);
      if (!aux) { /* doesn't exist locally */
          /* do nothing */
          ;
      }
      else { /* exists locally */
        if (mailptr_p->category == SET_INBOX)
          /* do nothing */
          ;
        else if (mailptr_p->category == SET_DELETED)
          aux->category = SET_DELETED;
      }
      mailptr_p = mailptr_p->next;
    }
  break;
  case PC2PILOT:
    mailptr_l = MailList_l;

    while(mailptr_l) {
      aux = js_Mail_list_lookup(MailList_p, mailptr_l->date);
      if (!aux) /* doesn't exist in pilot */
        js_Mail_list_insert(&MailList_p, mailptr_l);
      else { /* exists in pilot */
        if (aux->category == SET_INBOX)
          /* do nothing */
          ;
        else if (aux->category == SET_DELETED)
          aux->category = SET_INBOX;
      }
      mailptr_l = mailptr_l->next;
    }
  break;
  case BIDIREC:
    mailptr_l = MailList_l;
    mailptr_p = MailList_p;

    while (mailptr_p || mailptr_l) {
      flag = 0;
      if (mailptr_p) flag |= 0x10;
      if (mailptr_l) flag |= 0x01;

      if (flag==0x11) {
        if (mailptr_p->date > mailptr_l->date)
          flag = 0x01;
        else if (mailptr_p->date < mailptr_l->date)
          flag = 0x10;
        else /* = */
          flag = 0x11;
      }

      switch (flag) {
      case 0x11:
        if (mailptr_p->category == SET_INBOX) {
          if (js_Mail_list_compare(mailptr_p, mailptr_l)) {
            mailptr_l->id = 0;
            js_Mail_list_insert(&MailList_p, mailptr_l);
            mailptr_l = mailptr_l->next;
            continue;
          }
          else ; /* do nothing */
        }
        else if (mailptr_p->category == SET_DELETED) {
          if (js_Mail_list_compare(mailptr_p, mailptr_l)) {
            mailptr_p = mailptr_p->next;
            continue;
          }
          else
            mailptr_l->category = SET_DELETED;
        }
        mailptr_p = mailptr_p->next;
        mailptr_l = mailptr_l->next;
      break;
      case 0x10:
        mailptr_p = mailptr_p->next;
      break;
      case 0x01:
        if (mailptr_l->category == SET_INBOX)
          js_Mail_list_insert(&MailList_p, mailptr_l);
        mailptr_l = mailptr_l->next;
      break;
      } /*switch*/
    } /*while*/
  break;
  case DO_NOTHING:
  break;
  }
  return 0;
}




/* send to the Net all mails from pilot's outbox */
int send_mails()
{
  struct Mail_List *mailptr = MailList_p, copy_p, *ptr;

  while (mailptr != NULL) {
    struct Mail m;
    struct tm *mytm;
    time_t tt;

    if (mailptr->category != SET_OUTBOX) {
      mailptr = mailptr->next;
      continue;
    }

    m.subject = mailptr->subject;
    m.from = mailptr->from;

/* reset the date to <now> since pilot does not set m.dated */
    tt = time(0);
    mytm = localtime(&tt);
    m.date = *mytm;

    m.to = strdup(mailptr->to);
    m.cc = strdup(mailptr->cc); 
    m.bcc = strdup(mailptr->bcc);
    m.replyTo = strdup(mailptr->replyTo);
    m.sentTo = strdup(mailptr->sentTo);
    m.body = strdup(mailptr->body);
    mytm = localtime(&mailptr->date);
    m.date = *mytm;
    m.dated = 0;
    send_mail(pop_user, pop_host, &m);

/* reorder P list - due to new date */

/* copy to new element */
    copy_p.subject = strdup(mailptr->subject);
    copy_p.from = strdup(mailptr->from);
    copy_p.to = strdup(mailptr->to);
    copy_p.cc = strdup(mailptr->cc);
    copy_p.bcc = strdup(mailptr->bcc);
    copy_p.replyTo = strdup(mailptr->replyTo);
    copy_p.sentTo = strdup(mailptr->sentTo);
    copy_p.body = strdup(mailptr->body);
    copy_p.date = tt;
    copy_p.dated = 1;
    copy_p.popid = mailptr->popid;
    copy_p.id = mailptr->id;
    copy_p.category = SET_DELETED;
    copy_p.attr = mailptr->attr;

    ptr = mailptr->next; /* save position of next element */

/* remove incorrect element */
    js_Mail_list_remove(&MailList_p, mailptr);

/* insert element in correct place */
    js_Mail_list_insert(&MailList_p, &copy_p);

    mailptr = ptr; /* restore next position */
  }

  return 0;
}


/* POP retrieval of mail to L list */
int read_mails()
{
  struct Mail t;
  struct Mail_List elem;
  int ret, i;
  char str[1024];

  for (i=1;;i++) {
    ret=read_popmail(i, &t);

    if (ret == -2 || ret == -1)
      break;

    elem.id = 0;
    elem.popid = i;
    elem.attr = 0;
    elem.category = SET_INBOX;

    t.date.tm_sec = 0; /* pilot does not handle seconds */
    elem.dated = t.dated;
    elem.date = mktime(&t.date);
    elem.subject = t.subject;
    elem.from = t.from;
    if (!strchr(t.from,'@')) {
      if (t.from[strlen(t.from)-1] == '>') {
        t.from[strlen(t.from)-1] = 0;
        sprintf(str,"%s@%s>",t.from, pop_host);
      }
      else {
        sprintf(str,"%s@%s",t.from, pop_host);
      }
      elem.from = strdup(str);
    }
    elem.to = t.to;
    elem.cc = t.cc;
    elem.bcc = t.bcc;
    elem.replyTo = t.replyTo;
    elem.sentTo = t.sentTo;
    elem.body = t.body;

    js_Mail_list_insert(&MailList_l, &elem);
  }

  return 0;
}


/* use DELE pop command wherever needed */
int update_mails()
{
  struct Mail_List *mailptr = MailList_l;

  while (mailptr != NULL) {
    if (mailptr->category == SET_DELETED)
      dele_popmail(mailptr->popid);

    mailptr = mailptr->next;
  }

  return 0;
}


int get_mail_prefs()
{
  FILE *fp;
  char location[512];

  sprintf(location,"%s/%s/conf/jetmail.dat",home_dir,username);
  fp = fopen(location,"r");
  fgets(pop_host,256,fp); /* hostname */
  pop_host[strlen(pop_host)-1]=0;
  fgets(pop_user,20,fp); /* user */
  pop_user[strlen(pop_user)-1]=0;
  fgets(pop_pass,20,fp); /* pass */
  pop_pass[strlen(pop_pass)-1]=0;
  fclose(fp);

  return 0;
}



void js_MailExec(i, attr, mai, m, siz, category, recid)
int i;
int attr;
struct MailAppInfo mai;
struct Mail m;
int siz;
int category;
recordid_t recid;
{
  struct Mail_List elem;

  elem.id=recid;
  elem.popid=0;
  elem.attr=attr;
  elem.category=category;

  elem.dated = m.dated;
  elem.date = mktime(&m.date);

  if (m.subject)
    elem.subject = strdup(m.subject);
  else elem.subject = 0;
  if (m.from)
    elem.from = strdup(m.from);
  else elem.from = 0;
  if (m.to)
    elem.to = strdup(m.to);
  else elem.to = 0;
  if (m.cc)
    elem.cc = strdup(m.cc);
  else elem.cc = 0;
  if (m.bcc)
    elem.bcc = strdup(m.bcc);
  else elem.bcc = 0;
  if (m.replyTo)
    elem.replyTo = strdup(m.replyTo);
  else elem.replyTo = 0;
  if (m.sentTo)
    elem.sentTo = strdup(m.sentTo);
  else elem.sentTo = 0;
  if (m.body)
    elem.body = strdup(m.body);
  else elem.body = 0;

#ifdef __DEBUG
  printf("\tDate: %s (%d) Category: %d\nFrom: %s\n",
            ctime(&elem.date), elem.dated, category, m.from?m.from:"<>");
#endif

  js_Mail_list_insert(&MailList_p, &elem);
}

/* mode=1 => fastsynch=YES*/
int js_MailSynch(int sd, int mode, int synchType, char *path, char *logbuffer)
{
  int i, db, len, dlperr;
  unsigned char buffer[0xffff];
  char appblock[0xffff];
  struct MailAppInfo mai;
  struct DBInfo info;
  int count, code;
  char tcl_string[256];
  struct Mail_List *mailptr = NULL;
  recordid_t ret_id;
  int size_list_p = 0;

  MailList_l = MailList_p = NULL;

#ifdef __DEBUG
  printf("js_MailSynch [mode=%d synchtype='%c' path=%s]\n",
         mode, synchType, path);
#endif

  /* Open the database, store access handle in db */
  if ((dlperr=dlp_OpenDB(sd, 0, 0xC0, "MailDB", &db)) < 0) {
    strcat(logbuffer, "MailDB - Unable to open database.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

  if (js_GetDBInfo(sd, "MailDB", &info) < 0) {
    dlp_CloseDB(sd, db);
    return -1;
  }

  if ((dlperr=dlp_ReadAppBlock(sd, db, 0, (unsigned char *)appblock, 0xffff))
       < 0) {
    strcat(logbuffer, "MailDB - Unable to read app block.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

  unpack_MailAppInfo(&mai, (unsigned char *)appblock, 0xffff);

  if ((dlperr=dlp_ReadOpenDBInfo(sd, db, &count)) < 0) {
    strcat(logbuffer, "MailDB - Unable to open read DBInfo.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }


#ifdef __DEBUG
  printf("There are %d records\n", count);
  printf("Database modification date: %s\n",time_str(info.modifyDate));
  printf("      Database backup date: %s\n",time_str(info.backupDate));
#endif

  g_interp->result = (char *)Tcl_Alloc(4096);
  g_interp->freeProc = 0L;

/*POP*/
  get_mail_prefs();
  if (init_pop(pop_host, pop_user, pop_pass) < 0) {
    strcat(logbuffer, "Mail - Unable to connect to mail server\n");
    dlp_CloseDB(sd, db);
    return -1;
  }
  read_mails();

/*PILOT*/
  for (i=0;;i++) {
      struct Mail m;
      int attr,siz;
      int category;
      recordid_t recid;

      if (count > 0) {
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",i*100/count);
        code = Tcl_Eval(g_interp, tcl_string);
      }

      len = dlp_ReadRecordByIndex(sd, db, i, buffer, &recid,
                                      &siz,&attr, &category);

      if (len != dlpErrNotFound && len < 0) {
        strcat(logbuffer, "MailDB - Unable to read next record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

      if (len < 0) {
        break;
      }

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_Mail(&m, buffer, len);
      js_MailExec(i, attr, mai, m, siz, category, recid);
  } /* for */


/* SENDMAIL */
  if (synchType != PC2PILOT)
    send_mails(); /* send mails in pilot's outbox */

/*SYNC inbox + deleted */
  mail_sync_lists(synchType, mode);


/*WRITEBACK*/

#ifdef __DEBUG
  printf("Writing back to pilot...\n");
#endif

  mailptr = MailList_p;

  i = 0;
  size_list_p = js_Mail_list_count(MailList_p);

  while (mailptr) {
    struct Mail m;
    struct tm *mytm;

    sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (++i * 100 / (size_list_p > 0) ? size_list_p : i));
    code = Tcl_Eval(g_interp, tcl_string);

    if (mailptr->category == SET_DELETED)
      dlp_DeleteRecord(sd, db, 0, mailptr->id);

    if (mailptr->category != SET_INBOX) {
      mailptr = mailptr->next;
      continue;
    }

    m.subject = strdup(mailptr->subject);
    m.from = strdup(mailptr->from);
    m.to = strdup(mailptr->to);
    m.cc = strdup(mailptr->cc);
    m.bcc = strdup(mailptr->bcc);
    m.replyTo = strdup(mailptr->replyTo);
    m.sentTo = strdup(mailptr->sentTo);
    m.body = strdup(mailptr->body);
    m.dated = mailptr->dated;
    mytm = localtime(&mailptr->date);
    m.date = *mytm;

    len = pack_Mail(&m, buffer, 0xffff);

    if ((dlperr=dlp_WriteRecord(sd, db, mailptr->attr & dlpRecAttrSecret, mailptr->id,
      mailptr->category, buffer, len, &ret_id)) < 0) {
        strcat(logbuffer, "MailDB - Unable to write record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

    mailptr = mailptr->next;
  }

/* remove delete data back on PC */
  update_mails();
  close_pop();

  sprintf(tcl_string, "progress -2 %d \"Synchronizing\"", 200);
  code = Tcl_Eval(g_interp, tcl_string);

  if (synchType != PC2PILOT && synchType != DO_NOTHING)
    dlp_ResetSyncFlags(sd, db);
    
/* Close the database */
  dlp_CloseDB(sd, db);

  strcat(logbuffer, "OK Mail"); 
  switch (synchType) {
    case PC2PILOT:   strcat(logbuffer," Pilot<<-Desktop"); break;
    case PILOT2PC:   strcat(logbuffer," Pilot->>Desktop"); break;
    case BIDIREC:    strcat(logbuffer," Pilot<->Desktop"); break; 
  }
  strcat(logbuffer, "\n");

  js_Mail_list_free(&MailList_l);
  js_Mail_list_free(&MailList_p);

  return 0;
}
