#include <pi-memo.h>
#include "Synch.h"
#include "MemoSynch.h"
#include "Comm.h"
#include "Database.h"
#include "Handle.h"
#include "assert.h"
#include "memo-lists.h"
#include "defs.h"

#define SLOW_SYNCH 0
#define FAST_SYNCH 1

#define dlpRecAttrDelOrArch (dlpRecAttrDeleted | dlpRecAttrArchived)

extern Tcl_Interp *g_interp;
extern char home_dir[256];
extern char username[256];

struct Memo_List *MemoList_l;
struct Memo_List *MemoList_p;

int get_memo_location (char *location)
{
  FILE *fp;
  char str[80];

  sprintf(location,"%s/%s/conf/jetmemo.dat",home_dir,username);
  fp = fopen(location,"r");
  fgets(str,80,fp);
  fclose(fp);
  str[strlen(str)-1]=0;
  sprintf(location,"%s/memos", str);

  return 0;
}

void memo_ack(char *str, int i) {
#ifdef __DEBUG
  printf("%s (line %d)\n", str, i);
#endif
}

/* core routine for synch */
int memo_sync_lists(int synchType, int mode, char *logbuffer)
{
  struct Memo_List *memoptr_l, *memoptr_p, *aux;
  int flag;

  switch (synchType) {
  case PILOT2PC:
    memoptr_p = MemoList_p;
    while(memoptr_p) {
      aux = js_Memo_list_lookup(MemoList_l, memoptr_p->id);
      if (!aux) { /* doesn't exist locally */
        if (memoptr_p->attr & dlpRecAttrDelOrArch)
          /* nothing to do */
          ;
        else {
          /* new local record */
memo_ack("INS l",__LINE__);

          if (!(memoptr_p->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            memoptr_p = memoptr_p->next;
            continue;
          }

          memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Memo_list_insert(&MemoList_l, memoptr_p);
        }
      }
      else { /* exists locally */
        if (memoptr_p->attr & dlpRecAttrDelOrArch) {
          /* delete local record */
memo_ack("DELE l",__LINE__);
          js_Memo_list_remove(&MemoList_l, memoptr_p);
        }
        else {
          /* modify local record */
memo_ack("MOD l",__LINE__);
          memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Memo_list_modify(aux, memoptr_p);
        }
      }
      memoptr_p = memoptr_p->next;
    }
  break;
  case PC2PILOT:
    memoptr_l = MemoList_l;

    while(memoptr_l) {
      aux = js_Memo_list_lookup(MemoList_p, memoptr_l->id);
      if (!aux) { /* doesn't exist in pilot */
        if (memoptr_l->attr & dlpRecAttrDelOrArch) {
memo_ack("DEL l",__LINE__);
          aux = memoptr_l->next; /* remember position */
          /* delete local record */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_Memo_list_insert(&MemoList_p, memoptr_l);

          js_Memo_list_remove(&MemoList_l, memoptr_l);
          memoptr_l = aux;  /* restore pointer */

          continue;
        }
        else {
          if (!(memoptr_l->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            memoptr_l = memoptr_l->next;
            continue;
          }

          /* new pilot record */
printf("inserting p\n");
          memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Memo_list_insert(&MemoList_p, memoptr_l);
        }
      }
      else { /* exists in pilot */
        if (memoptr_l->attr & dlpRecAttrDelOrArch) {
          /* delete pilot record */
printf("deleting l, is in pilot\n");
          aux = memoptr_l->next; /* remember position */
          /* delete local record */
          js_Memo_list_remove(&MemoList_l, memoptr_l);
          memoptr_l = aux;  /* restore pointer */
          continue;
        }
        else {
          /* modify pilot record */
printf("modifying p\n");
          memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Memo_list_modify(aux, memoptr_l);
        }
      }
      memoptr_l = memoptr_l->next;
    }
  break;
  case BIDIREC:
    memoptr_l = MemoList_l;
    memoptr_p = MemoList_p;
    while (memoptr_p || memoptr_l) {
      flag = 0;
      if (memoptr_p) flag |= 0x10;
      if (memoptr_l) flag |= 0x01;

      if (flag==0x11) {
        if (memoptr_p->id > memoptr_l->id)
          flag = 0x01;
        else if (memoptr_p->id < memoptr_l->id)
          flag = 0x10;
        else /* = */
          flag = 0x11;
      }

      switch (flag) {
      case 0x11:
        if (memoptr_p->attr & dlpRecAttrDelOrArch) {
          if (memoptr_l->attr & dlpRecAttrDelOrArch) {
memo_ack("P:D L:D",__LINE__);
            aux = memoptr_l->next; /* remember position */
            js_Memo_list_remove(&MemoList_l, memoptr_l);
            memoptr_l = aux;  /* restore pointer */
        
            memoptr_p = memoptr_p->next;
          }
          else if (memoptr_l->attr & dlpRecAttrDirty) {
memo_ack("P:D L:Y",__LINE__);
            memoptr_p->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_Memo_list_modify(memoptr_p, memoptr_l);

            memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            memoptr_p = memoptr_p->next;
            memoptr_l = memoptr_l->next;
          }
          else {
memo_ack("P:D L:-",__LINE__);
            aux = memoptr_l->next; /* remember position */
            js_Memo_list_remove(&MemoList_l, memoptr_l);
            memoptr_l = aux;  /* restore pointer */

            memoptr_p = memoptr_p->next;
          }
        }
        else if (memoptr_p->attr & dlpRecAttrDirty) {
          if (memoptr_l->attr & dlpRecAttrDelOrArch) {
memo_ack("P:Y L:D",__LINE__);
            memoptr_l->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_Memo_list_modify(memoptr_l, memoptr_p);

            memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            memoptr_p = memoptr_p->next;
            memoptr_l = memoptr_l->next;
          }
          else if (memoptr_l->attr & dlpRecAttrDirty) {
memo_ack("P:Y L:Y",__LINE__);
            if (js_Memo_list_compare(memoptr_p, memoptr_l)) {
              memoptr_l->id = 0;
  
              memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
              memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
  
              js_Memo_list_insert(&MemoList_p, memoptr_l);
              js_Memo_list_insert(&MemoList_l, memoptr_l);
              memoptr_l->id = memoptr_p->id;
              js_Memo_list_modify(memoptr_l, memoptr_p);
              strcat(logbuffer,
              "Duplicating record since it was changed on both sides.\n");
            }
            
            memoptr_p = memoptr_p->next;
            memoptr_l = memoptr_l->next;
          }
          else {
memo_ack("P:Y L:-",__LINE__);
            memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_Memo_list_modify(memoptr_l, memoptr_p);

            memoptr_p = memoptr_p->next;
            memoptr_l = memoptr_l->next;
          }
        }
        else {
          if (memoptr_l->attr & dlpRecAttrDelOrArch) {
memo_ack("P:- L:D",__LINE__);
            memoptr_p->attr |= dlpRecAttrDelOrArch;  /* set deleted flag */

            aux = memoptr_l->next; /* remember position */
            js_Memo_list_remove(&MemoList_l, memoptr_l);
            memoptr_l = aux;  /* restore pointer */

            memoptr_p = memoptr_p->next;
          }
          else if (memoptr_l->attr & dlpRecAttrDirty) {
memo_ack("P:- L:Y",__LINE__);
            memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_Memo_list_modify(memoptr_p, memoptr_l);

            memoptr_p = memoptr_p->next;
            memoptr_l = memoptr_l->next;
          }
          else {
memo_ack("P:- L:-",__LINE__);
/* nothing to do */
            if (0)

/* uncomment this code and comment the above if, in case records should be
   duplicated when a slow synch occurs (records were modified with
   different PC's)
*/
/*            if (mode != FAST_SYNCH)*/
              if (js_Memo_list_compare(memoptr_p, memoptr_l)) {
                memoptr_l->id = 0;
  
                memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
                memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
  
                js_Memo_list_insert(&MemoList_p, memoptr_l);
                js_Memo_list_insert(&MemoList_l, memoptr_l);
                memoptr_l->id = memoptr_p->id;
                js_Memo_list_modify(memoptr_l, memoptr_p);
                strcat(logbuffer,
                "Duplicating record since a slow synch occured and data was changed.\n"); 
              }

            memoptr_p = memoptr_p->next;
            memoptr_l = memoptr_l->next;
          }
        }
      break;
      case 0x10:
        if (memoptr_p->attr & dlpRecAttrDelOrArch)
memo_ack("0x10 P:D",__LINE__)
          /* nothing to do */
          ;
        else if (memoptr_p->attr & dlpRecAttrDirty) {
memo_ack("0x10 P:Y",__LINE__);
          memoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Memo_list_insert(&MemoList_l, memoptr_p);
        }
        else {
memo_ack("0x10 P:-",__LINE__);
          js_Memo_list_insert(&MemoList_l, memoptr_p);
        }

        memoptr_p = memoptr_p->next;
      break;
      case 0x01:
        if (memoptr_l->attr & dlpRecAttrDelOrArch) {
memo_ack("0x01 L:D",__LINE__);
          aux = memoptr_l->next; /* remember position */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_Memo_list_insert(&MemoList_p, memoptr_l);
          js_Memo_list_remove(&MemoList_l, memoptr_l);
          memoptr_l = aux;  /* restore pointer */
          continue;
        }
        else if (memoptr_l->attr & dlpRecAttrDirty) {
memo_ack("0x01 L:Y",__LINE__);
          memoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_Memo_list_insert(&MemoList_p, memoptr_l);
        }
        else {
memo_ack("0x01 L:-",__LINE__);
          if (mode != FAST_SYNCH)
            js_Memo_list_insert(&MemoList_p, memoptr_l);
        }

        memoptr_l = memoptr_l->next;
      break;
      } /*switch*/
    } /*while*/
  break;
  case DO_NOTHING:
  break;
  }
  return 0;
}

int write_memos()
{
  char location[512];
  FILE *fp;
  struct Memo_List *memoptr = MemoList_l;

  get_memo_location(location);

  if (!(fp = fopen(location,"w"))) {
    printf("Couldn't open %s for writing...\n",location);
    return -1; 
  }
  while (memoptr != NULL) {
    fprintf(fp,"%ld %d %d %d\n%s\n",
            memoptr->id, memoptr->size, memoptr->category,
            memoptr->attr, memoptr->data); /* write record */

    memoptr = memoptr->next;
  }
  fclose(fp);

  return 0;
}


int read_memos()
{
  char location[512];
  FILE *fp;
  char str[80];
  int count=0;
  struct Memo_List elem;

  get_memo_location(location);

  if (!(fp = fopen(location,"r"))) {
    printf("Couldn't open %s for reading...\n",location);
    return -1; 
  }
  while (1) {
    fgets(str,80,fp); /* read header */

    if (feof(fp)) break;

/* read memo header*/
    sscanf(str,"%ld %d %d %d",&elem.id,&elem.size,&elem.category,&elem.attr);

/* read memo body */
    elem.data = (char *)malloc(elem.size+1);
    fread(elem.data,elem.size+1,1,fp); /* read memo in record */
    if (ferror(fp)) /* read memo in record */
      printf("Error in fread (line %d, file %s)\n",__LINE__,__FILE__);
    elem.data[elem.size]=0;

/* add memo to list */
    js_Memo_list_insert(&MemoList_l, &elem);
    if (elem.data) free(elem.data);
    count++;
  }

  fclose(fp);

  return 0;
}


void memo_save_categories(struct MemoAppInfo mai)
{
  FILE *fp;
  char location[256];
  int i;

  sprintf(location,"%s/%s/conf/jetmemo.cat", home_dir, username);
  fp = fopen(location,"w");
  for (i = 0; i < 16; i++) {
    fprintf(fp,"%s\n",mai.category.name[i]);
  }
  fclose(fp);
}

void js_MemoExec(i, attr, mai, m, siz, category, recid)
int i;
int attr;
struct MemoAppInfo mai;
struct Memo m;
int siz;
int category;
recordid_t recid;
{
/*** Insert code here !! ***/

/*** Example: ***/
#ifdef __DEBUG
  char *ptr;
#endif
  struct Memo_List elem;

  elem.id=recid;
  elem.attr=attr;
  elem.category=category;

  elem.size=siz;
  if (!(attr & dlpRecAttrDelOrArch)) {
/*  if (siz) {*/   /* if using archive uncomment this line */
    --elem.size;
    elem.data = (char *) malloc(siz + 1);
    strcpy(elem.data, m.text);

#ifdef __DEBUG
    printf("\tSize: %d  Category: %s  ",
            siz, mai.category.name[category]);
    ptr=m.text;
    while (*ptr && *ptr != '\n') ptr++;
    *ptr=0;
    printf("Contents: \"%s[...]\"\n",m.text);
#endif
  } else {
    elem.size = 0;
  }

  js_Memo_list_insert(&MemoList_p, &elem);
}

/* mode=1 => fastsynch=YES*/
int js_MemoSynch(int sd, int mode, int synchType, char *path, char *logbuffer)
{
  int i, db, len, dlperr;
  unsigned char buffer[0xffff];
  char appblock[0xffff];
  struct MemoAppInfo mai;
  struct DBInfo info;
  int count, cnt=0, code;
  char tcl_string[256];
  struct Memo_List *memoptr = NULL, *aux;
  recordid_t ret_id;
  int size_list_p = 0;

  MemoList_l = MemoList_p = NULL;

#ifdef __DEBUG
  printf("js_MemoSynch [mode=%d synchtype='%c' path=%s]\n",
         mode, synchType, path);
#endif

  /* Open the database, store access handle in db */
  if ((dlperr=dlp_OpenDB(sd, 0, 0xC0, "MemoDB", &db)) < 0) {
    strcat(logbuffer, "MemoDB - Unable to open database.\n");
    return -1;
  }

  if (js_GetDBInfo(sd, "MemoDB", &info) < 0) {
    dlp_CloseDB(sd, db);
    return -1;
  }

  if ((dlperr=dlp_ReadAppBlock(sd, db, 0, (unsigned char *)appblock, 0xffff))
       < 0) {
    strcat(logbuffer, "MemoDB - Unable to read app block.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

  unpack_MemoAppInfo(&mai, (unsigned char *)appblock, 0xffff);

  memo_save_categories(mai);

  if ((dlperr=dlp_ReadOpenDBInfo(sd, db, &count)) < 0) {
    strcat(logbuffer, "MemoDB - Unable to open read DBInfo.\n");
    dlp_CloseDB(sd, db);
    return -1;
  }

#ifdef __DEBUG
  printf("There are %d records\n", count);
printf("Database modification date: %s\n",time_str(info.modifyDate));
printf("      Database backup date: %s\n",time_str(info.backupDate));
#endif

  g_interp->result = (char *)Tcl_Alloc(4096);
  g_interp->freeProc = 0L;

/* read local memos from PC */
  read_memos();

  if (mode == SLOW_SYNCH) {
    for (i=0;;i++) {
      struct Memo m;
      int attr,siz;
      int category;
      recordid_t recid;

      if (count > 0) {
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",i*100/count);
        code = Tcl_Eval(g_interp, tcl_string);
      }

      len = dlp_ReadRecordByIndex(sd, db, i, buffer, &recid,
                                      &siz,&attr, &category);

      if (len != dlpErrNotFound && len < 0) {
        strcat(logbuffer, "MemoDB - Unable to read next record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

      if (len < 0) {
        break;
      }

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_Memo(&m, buffer, len);
      js_MemoExec(i, attr, mai, m, siz, category, recid);
    } /* for */
  }
  else if (mode == FAST_SYNCH) {
    int last_modnum = get_mod_num();
    /* algoritmo heuristico para descodificacao do numero de
       records modificados -- estimativa razoavelmente aproximada */
    count = (int) ((info.modnum - last_modnum + 2) / 5);

    for (i=0;;i++) {
      struct Memo m;
      int attr,siz,j;
      int category;
      recordid_t recid;

      if (count)
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
                i/count?100:i*100/count);
      else
        sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");

      code = Tcl_Eval(g_interp, tcl_string);

      len = dlp_ReadNextModifiedRec(sd, db, buffer, &recid, &j,
                                      &siz,&attr, &category);

      if (len != dlpErrNotFound && len < 0) {
        strcat(logbuffer, "MemoDB - Unable to read next modified record.\n");
        dlp_CloseDB(sd, db);
        return -1;
      }

      if (len < 0) { /* no more records to read */
        count=cnt;

        sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");
        code = Tcl_Eval(g_interp, tcl_string);
        break;
      }

      cnt++;

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_Memo(&m, buffer, len);
      js_MemoExec(i, attr, mai, m, siz, category, recid);
    } /* for */
  } /* fast / slow */


/*SYNC*/
  memo_sync_lists(synchType, mode, logbuffer);


/*WRITEBACK*/

  save_mod_num(info.modnum+count);

#ifdef __DEBUG
  printf("Writing back to pilot...\n");
#endif

  memoptr = MemoList_p;

  aux = MemoList_l;
  i = 0;
  size_list_p = js_Memo_list_count(MemoList_p);

  while (memoptr) {
    struct Memo m;

    i++;
    if (size_list_p > 0)
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / size_list_p));
    else
      sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / i));

    code = Tcl_Eval(g_interp, tcl_string);

    /* Skip deleted records */
    if (memoptr->attr & dlpRecAttrDelOrArch) {

      if ((dlperr=dlp_DeleteRecord(sd, db, 0, memoptr->id)) < 0) {
        strcat(logbuffer, "MemoDB - Unable to delete record.\n");
        memoptr = memoptr->next;
        continue;
      }

#ifdef __DEBUG
      printf("Deleting record %ld\n",memoptr->id);
#endif
      memoptr = memoptr->next;
      continue;
    }

    /* Skip dirty records */
    if (memoptr->attr & dlpRecAttrDirty) {
      memoptr = memoptr->next;
      continue;
    }

    m.text = strdup(memoptr->data);
    pack_Memo(&m, buffer, 0xffff);
    len = -1;  /* len should be the value returned by pack_Memo() */

    if ((dlperr=dlp_WriteRecord(sd, db, memoptr->attr & dlpRecAttrSecret, memoptr->id,
      memoptr->category, buffer, len, &ret_id)) < 0) {
        strcat(logbuffer, "MemoDB - Unable to write record.\n");
        dlp_CloseDB(sd, db);
        return -1;
    }


    if (memoptr->id == 0) { /* pilot returns new record's id */
      aux->id = ret_id;
      aux=aux->next;
    }

    memoptr = memoptr->next;
  }

/* write local data back to PC */
  write_memos();

  sprintf(tcl_string, "progress -2 %d \"Synchronizing\"", 200);
  code = Tcl_Eval(g_interp, tcl_string);

  if (synchType != PC2PILOT && synchType != DO_NOTHING)
    dlp_ResetSyncFlags(sd, db);
    
/* Close the database */
  dlp_CloseDB(sd, db);

  strcat(logbuffer, "OK Memo");
  switch (synchType) {
    case PC2PILOT:   strcat(logbuffer," Pilot<<-Desktop"); break;
    case PILOT2PC:   strcat(logbuffer," Pilot->>Desktop"); break;
    case BIDIREC:    strcat(logbuffer," Pilot<->Desktop"); break;
  }
  strcat(logbuffer, "\n");
  js_Memo_list_free(&MemoList_l);
  js_Memo_list_free(&MemoList_p);

  return 0;
}
