#include "Synch.h"
#include "ToDoSynch.h"
#include <pi-todo.h>
#include "Comm.h"
#include "Database.h"
#include "Handle.h"
#include "assert.h"
#include "todo-lists.h"
#include "defs.h"

#define SLOW_SYNCH 0
#define FAST_SYNCH 1

#define dlpRecAttrDelOrArch (dlpRecAttrDeleted | dlpRecAttrArchived)

extern Tcl_Interp *g_interp;
extern char home_dir[256];
extern char username[256];

struct ToDo_List *ToDoList_l;
struct ToDo_List *ToDoList_p;

int get_todo_location (char *location)
{
  FILE *fp;
  char str[80];

  sprintf(location,"%s/%s/conf/jettodo.dat",home_dir,username);
  fp = fopen(location,"r");
  fgets(str,80,fp);
  fclose(fp);
  str[strlen(str)-1]=0;
  sprintf(location,"%s/todos", str);

  return 0;
}

void ack(char *str, int i) {
#ifdef __DEBUG
  printf("%s (line %d)\n", str, i);
#endif
}

/* core routine for synch */
int sync_lists(int synchType, int mode, char *logbuffer)
{
  struct ToDo_List *todoptr_l, *todoptr_p, *aux;
  int flag;

  switch (synchType) {
  case PILOT2PC:
    todoptr_p = ToDoList_p;
    while(todoptr_p) {
      aux = js_ToDo_list_lookup(ToDoList_l, todoptr_p->id);
      if (!aux) { /* doesn't exist locally */
        if (todoptr_p->attr & dlpRecAttrDelOrArch)
          /* nothing to do */
          ;
        else {
          /* new local record */
ack("INS l",__LINE__);

          if (!(todoptr_p->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            todoptr_p = todoptr_p->next;
            continue;
          }

          todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_ToDo_list_insert(&ToDoList_l, todoptr_p);
        }
      }
      else { /* exists locally */
        if (todoptr_p->attr & dlpRecAttrDelOrArch) {
          /* delete local record */
ack("DELE l",__LINE__);
          js_ToDo_list_remove(&ToDoList_l, todoptr_p);
        }
        else {
          /* modify local record */
ack("MOD l",__LINE__);
          todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_ToDo_list_modify(aux, todoptr_p);
        }
      }
      todoptr_p = todoptr_p->next;
    }
  break;
  case PC2PILOT:
    todoptr_l = ToDoList_l;

    while(todoptr_l) {
      aux = js_ToDo_list_lookup(ToDoList_p, todoptr_l->id);
      if (!aux) { /* doesn't exist in pilot */
        if (todoptr_l->attr & dlpRecAttrDelOrArch) {
ack("DEL l",__LINE__);
          aux = todoptr_l->next; /* remember position */
          /* delete local record */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_ToDo_list_insert(&ToDoList_p, todoptr_l);

          js_ToDo_list_remove(&ToDoList_l, todoptr_l);
          todoptr_l = aux;  /* restore pointer */

          continue;
        }
        else {
          if (!(todoptr_l->attr & dlpRecAttrDirty) && mode==FAST_SYNCH) {
            todoptr_l = todoptr_l->next;
            continue;
          }

          /* new pilot record */
printf("inserting p\n");
          todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_ToDo_list_insert(&ToDoList_p, todoptr_l);
        }
      }
      else { /* exists in pilot */
        if (todoptr_l->attr & dlpRecAttrDelOrArch) {
          /* delete pilot record */
printf("deleting l, is in pilot\n");
          aux = todoptr_l->next; /* remember position */
          /* delete local record */
          js_ToDo_list_remove(&ToDoList_l, todoptr_l);
          todoptr_l = aux;  /* restore pointer */
          continue;
        }
        else {
          /* modify pilot record */
printf("modifying p\n");
          todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_ToDo_list_modify(aux, todoptr_l);
        }
      }
      todoptr_l = todoptr_l->next;
    }
  break;
  case BIDIREC:
    todoptr_l = ToDoList_l;
    todoptr_p = ToDoList_p;
    while (todoptr_p || todoptr_l) {
      flag = 0;
      if (todoptr_p) flag |= 0x10;
      if (todoptr_l) flag |= 0x01;

      if (flag==0x11) {
        if (todoptr_p->id > todoptr_l->id)
          flag = 0x01;
        else if (todoptr_p->id < todoptr_l->id)
          flag = 0x10;
        else /* = */
          flag = 0x11;
      }

      switch (flag) {
      case 0x11:
        if (todoptr_p->attr & dlpRecAttrDelOrArch) {
          if (todoptr_l->attr & dlpRecAttrDelOrArch) {
ack("P:D L:D",__LINE__);
            aux = todoptr_l->next; /* remember position */
            js_ToDo_list_remove(&ToDoList_l, todoptr_l);
            todoptr_l = aux;  /* restore pointer */
        
            todoptr_p = todoptr_p->next;
          }
          else if (todoptr_l->attr & dlpRecAttrDirty) {
ack("P:D L:Y",__LINE__);
            todoptr_p->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_ToDo_list_modify(todoptr_p, todoptr_l);

            todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            todoptr_p = todoptr_p->next;
            todoptr_l = todoptr_l->next;
          }
          else {
ack("P:D L:-",__LINE__);
            aux = todoptr_l->next; /* remember position */
            js_ToDo_list_remove(&ToDoList_l, todoptr_l);
            todoptr_l = aux;  /* restore pointer */

            todoptr_p = todoptr_p->next;
          }
        }
        else if (todoptr_p->attr & dlpRecAttrDirty) {
          if (todoptr_l->attr & dlpRecAttrDelOrArch) {
ack("P:Y L:D",__LINE__);
            todoptr_l->attr &= ~dlpRecAttrDelOrArch;  /* unset deleted flag */
            js_ToDo_list_modify(todoptr_l, todoptr_p);

            todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

            todoptr_p = todoptr_p->next;
            todoptr_l = todoptr_l->next;
          }
          else if (todoptr_l->attr & dlpRecAttrDirty) {
ack("P:Y L:Y",__LINE__);
            if (js_ToDo_list_compare(todoptr_p, todoptr_l)) {
              todoptr_l->id = 0;

              todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
              todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */

              js_ToDo_list_insert(&ToDoList_p, todoptr_l);
              js_ToDo_list_insert(&ToDoList_l, todoptr_l);
              todoptr_l->id = todoptr_p->id;
              js_ToDo_list_modify(todoptr_l, todoptr_p);
              strcat(logbuffer,
              "Duplicating record since it was changed on both sides.\n"); 
            }
            
            todoptr_p = todoptr_p->next;
            todoptr_l = todoptr_l->next;
          }
          else {
ack("P:Y L:-",__LINE__);
            todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_ToDo_list_modify(todoptr_l, todoptr_p);

            todoptr_p = todoptr_p->next;
            todoptr_l = todoptr_l->next;
          }
        }
        else {
          if (todoptr_l->attr & dlpRecAttrDelOrArch) {
ack("P:- L:D",__LINE__);
            todoptr_p->attr |= dlpRecAttrDelOrArch;  /* set deleted flag */

            aux = todoptr_l->next; /* remember position */
            js_ToDo_list_remove(&ToDoList_l, todoptr_l);
            todoptr_l = aux;  /* restore pointer */

            todoptr_p = todoptr_p->next;
          }
          else if (todoptr_l->attr & dlpRecAttrDirty) {
ack("P:- L:Y",__LINE__);
            todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset deleted flag */
            js_ToDo_list_modify(todoptr_p, todoptr_l);

            todoptr_p = todoptr_p->next;
            todoptr_l = todoptr_l->next;
          }
          else {
ack("P:- L:-",__LINE__);
/* nothing to do */
            if (0)                 

/* uncomment this code and comment the above if, in case records should be
   duplicated when a slow synch occurs (records were modified with
   different PC's)
*/
/*            if (mode != FAST_SYNCH)*/
              if (js_ToDo_list_compare(todoptr_p, todoptr_l)) {
                todoptr_l->id = 0;
                
                todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
                todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
                
                js_ToDo_list_insert(&ToDoList_p, todoptr_l);
                js_ToDo_list_insert(&ToDoList_l, todoptr_l);
                todoptr_l->id = todoptr_p->id;
                js_ToDo_list_modify(todoptr_l, todoptr_p);
                strcat(logbuffer,
                "Duplicating record since a slow synch occured and data was changed.\n");
              }

            todoptr_p = todoptr_p->next;
            todoptr_l = todoptr_l->next;
          }
        }
      break;
      case 0x10:
        if (todoptr_p->attr & dlpRecAttrDelOrArch)
ack("0x10 P:D",__LINE__)
          /* nothing to do */
          ;
        else if (todoptr_p->attr & dlpRecAttrDirty) {
ack("0x10 P:Y",__LINE__);
          todoptr_p->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_ToDo_list_insert(&ToDoList_l, todoptr_p);
        }
        else {
ack("0x10 P:-",__LINE__);
          js_ToDo_list_insert(&ToDoList_l, todoptr_p);
        }

        todoptr_p = todoptr_p->next;
      break;
      case 0x01:
        if (todoptr_l->attr & dlpRecAttrDelOrArch) {
ack("0x01 L:D",__LINE__);
          aux = todoptr_l->next; /* remember position */

/* if fast sync then insert into P list */
          if (mode == FAST_SYNCH)
            js_ToDo_list_insert(&ToDoList_p, todoptr_l);
          js_ToDo_list_remove(&ToDoList_l, todoptr_l);
          todoptr_l = aux;  /* restore pointer */
          continue;
        }
        else if (todoptr_l->attr & dlpRecAttrDirty) {
ack("0x01 L:Y",__LINE__);
          todoptr_l->attr &= ~dlpRecAttrDirty;  /* unset dirty flag */
          js_ToDo_list_insert(&ToDoList_p, todoptr_l);
        }
        else {
ack("0x01 L:-",__LINE__);
          if (mode != FAST_SYNCH)
            js_ToDo_list_insert(&ToDoList_p, todoptr_l);
        }

        todoptr_l = todoptr_l->next;
      break;
      } /*switch*/
    } /*while*/
  break;
  case DO_NOTHING:
  break;
  }
  return 0;
}

int write_todos()
{
  char location[512], str[15];
  FILE *fp;
  struct ToDo_List *todoptr = ToDoList_l;

  get_todo_location(location);
     
  if (!(fp = fopen(location,"w"))) {
    printf("Couldn't open %s for writing...\n",location);
    return -1;
  }

  while (todoptr != NULL) {
    if (strcmp(todoptr->duedate,"NODATE")) {
      sprintf(str,"%c%c%c%c%c%c%c%c",
        todoptr->duedate[0],
        todoptr->duedate[1],

        todoptr->duedate[3],
        todoptr->duedate[4],

        todoptr->duedate[6],
        todoptr->duedate[7],
        todoptr->duedate[8],
        todoptr->duedate[9]);
      strcpy(todoptr->duedate, str);
    }
    
    /* write record */
    fprintf(fp,"%ld %d %d %d %d %d %d %s\n%s\n%s\n",
            todoptr->id, todoptr->check, todoptr->category,
            todoptr->attr, todoptr->priority,
            todoptr->size_n, todoptr->size_d, todoptr->duedate,
            todoptr->note, todoptr->desc);
    todoptr = todoptr->next;
  }
  fclose(fp);
        
  return 0;
}


int read_todos()
{ 
  char location[512];
  FILE *fp;
  char str[1024], auxstr[20];
  struct ToDo_List elem;
  
  get_todo_location(location);
  if (!(fp = fopen(location,"r"))) {
    printf("Couldn't open %s for reading...\n",location);
    return -1;
  }
  while (1) {
    fgets(str,1024,fp); /* read header */

    if (feof(fp)) break;

/* read todo header*/ 
    sscanf(str,"%ld %d %d %d %d %d %d %s\n",
           &elem.id,&elem.check, &elem.category,&elem.attr,
           &elem.priority, &elem.size_n, &elem.size_d, auxstr);

  
    if (!strcmp(auxstr,"NODATE"))
      strcpy(elem.duedate,"NODATE");
    else
      sprintf(elem.duedate,"%c%c/%c%c/%s", auxstr[0], auxstr[1],
        auxstr[2], auxstr[3], &auxstr[4]);

/* read todo body */
    elem.note = (char *)malloc(elem.size_n+1);
    elem.desc = (char *)malloc(elem.size_d+1);
                    
    fread(elem.note,elem.size_n+1,1,fp); /* read todo note */
    elem.note[elem.size_n]=0;
    fread(elem.desc,elem.size_d+1,1,fp); /* read todo desc */
    elem.desc[elem.size_d]=0;

/* add todo to list */
    js_ToDo_list_insert(&ToDoList_l, &elem);
    if (elem.note) free(elem.note);
    if (elem.desc) free(elem.desc);
  }
  fclose(fp);

  return 0;
}


void todo_save_categories(struct ToDoAppInfo mai)
{
  FILE *fp;
  char location[256];
  int i;
  
  sprintf(location,"%s/%s/conf/jettodo.cat", home_dir, username);
  fp = fopen(location,"w");
  for (i = 0; i < 16; i++) {
    fprintf(fp,"%s\n",mai.category.name[i]);
  }
  fclose(fp);
}


void js_ToDoExec(i, attr, mai, m, siz, category, recid)
int i;
int attr;
struct ToDoAppInfo mai;
struct ToDo m;
int siz;
int category;
recordid_t recid;
{
  struct ToDo_List elem;

  elem.attr=attr;
  elem.category=category;
  elem.id=recid;

  elem.check=m.complete;
  elem.priority=m.priority;
  if (m.indefinite)
    strcpy(elem.duedate, "NODATE");
  else
    strftime(elem.duedate, 15, "%d/%m/%Y", &m.due);

  if (!(attr & dlpRecAttrDelOrArch)) {
    if (m.note) {
      elem.size_n = strlen(m.note);
      elem.note = strdup(m.note);
    }
    else
      elem.size_n = 0;

    if (m.description) {
      elem.size_d = strlen(m.description);
      elem.desc = strdup(m.description);
    }
    else
      elem.size_d = 0;

#ifdef __DEBUG
    printf("\tSize: %d  Category: %s  ",
            siz, mai.category.name[category]);

    printf("Contents: \"%s\"\n",m.description?m.description:"<<EMPTY>>");
    printf("Note:\n'%s'\n",m.note?m.note:"<<EMPTY>>");
#endif

  } else {
    elem.size_d = elem.size_n = 0;
  }

  js_ToDo_list_insert(&ToDoList_p, &elem);
}

/* mode=1 => fastsynch=YES*/
int js_ToDoSynch(int sd, int mode, int synchType, char *path, char *logbuffer)
{
  int i, db;
  unsigned char buffer[0xffff];
  int len;
  char appblock[0xffff];
  struct ToDoAppInfo mai;
  struct DBInfo info;
  int count, cnt=0, code;
  char tcl_string[256];
  struct ToDo_List *todoptr = NULL, *aux;
  recordid_t ret_id;
  int size_list_p = 0;

  ToDoList_l = ToDoList_p = NULL;

#ifdef __DEBUG
  printf("js_ToDoSynch [mode=%d synchtype='%c' path=%s]\n",
         mode, synchType, path);
#endif

  /* Open the database, store access handle in db */
  if (dlp_OpenDB(sd, 0, 0xC0, "ToDoDB", &db) < 0) {
    puts("Unable to open database");
    dlp_AddSyncLogEntry(sd, "Unable to open database.\n");
    return -1;
  }

  js_GetDBInfo(sd, "ToDoDB", &info);

  dlp_ReadAppBlock(sd, db, 0, (unsigned char *)appblock, 0xffff);

  unpack_ToDoAppInfo(&mai, (unsigned char *)appblock, 0xffff);

  todo_save_categories(mai);

  dlp_ReadOpenDBInfo(sd, db, &count);


#ifdef __DEBUG
  printf("There are %d records\n", count);
  printf("Database modification date: %s\n",time_str(info.modifyDate));
  printf("      Database backup date: %s\n",time_str(info.backupDate));
#endif

  g_interp->result = (char *)Tcl_Alloc(4096);
  g_interp->freeProc = 0L;

/* read local todos from PC */
  read_todos();

  if (mode == SLOW_SYNCH) {
    for (i=0;;i++) {
      struct ToDo m;
      int attr,siz;
      int category;
      recordid_t recid;

      if (count > 0) {
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",i*100/count);
        code = Tcl_Eval(g_interp, tcl_string);
      }

      len = dlp_ReadRecordByIndex(sd, db, i, buffer, &recid,
                                      &siz,&attr, &category);

      if (len < 0)
        break;

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_ToDo(&m, buffer, len);
      js_ToDoExec(i, attr, mai, m, siz, category, recid);
    } /* for */
  }
  else if (mode == FAST_SYNCH) {
    int last_modnum = get_mod_num();
    /* algoritmo heuristico para descodificacao do numero de
       records modificados -- estimativa razoavelmente aproximada */
    count = (int) ((info.modnum - last_modnum + 2) / 5);

    for (i=0;;i++) {
      struct ToDo m;
      int attr,siz,j;
      int category;
      recordid_t recid;

      if (count)
        sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
                i/count?100:i*100/count);
      else
        sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");

      code = Tcl_Eval(g_interp, tcl_string);

      len = dlp_ReadNextModifiedRec(sd, db, buffer, &recid, &j,
                                      &siz,&attr, &category);

      if (len < 0) { /* no more records to read */
        count=cnt;

        sprintf(tcl_string, "progress -2 100 \"Synchronizing\"");
        code = Tcl_Eval(g_interp, tcl_string);
        break;
      }

      cnt++;

#ifdef __DEBUG
      printf("Record %d (RECID=%ld) [%s%s%s%s%s]\n", i, recid,
            attr & dlpRecAttrDeleted?"D":"-",
            attr & dlpRecAttrArchived?"A":"-",
            attr & dlpRecAttrBusy?"B":"-",
            attr & dlpRecAttrSecret?"S":"-",
            attr & dlpRecAttrDirty?"Y":"-"
            );
#endif

      unpack_ToDo(&m, buffer, len);
      js_ToDoExec(i, attr, mai, m, siz, category, recid);
    } /* for */
  } /* fast / slow */


/*SYNC*/
  sync_lists(synchType, mode, logbuffer);


/*WRITEBACK*/

  save_mod_num(info.modnum+count);

#ifdef __DEBUG
  printf("Writing back to pilot...\n");
#endif

  todoptr = ToDoList_p;

  aux = ToDoList_l;
  i = 0;
  size_list_p = js_ToDo_list_count(ToDoList_p);
  while (todoptr) {
    struct ToDo m;

  i++;
  if (size_list_p > 0)
    sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / size_list_p));
  else
    sprintf(tcl_string, "progress -2 %d \"Synchronizing\"",
            100 + (i * 100 / i));

    code = Tcl_Eval(g_interp, tcl_string);

    /* Skip deleted records */
    if (todoptr->attr & dlpRecAttrDelOrArch) {
      dlp_DeleteRecord(sd, db, 0, todoptr->id);
#ifdef __DEBUG
      printf("Deleting record %ld\n",todoptr->id);
#endif
      todoptr = todoptr->next;
      continue;
    }

    /* Skip dirty records */  
    if (todoptr->attr & dlpRecAttrDirty) {
      todoptr = todoptr->next;
      continue;
    }

/* handle strings */
    if (todoptr->size_d && todoptr->desc[0]) {
      m.description = strdup(todoptr->desc);
    }
    else
      m.description = "";

    if (todoptr->size_n && todoptr->note[0]) {
      m.note = strdup(todoptr->note);
    }
    else
      m.note = "";


/* handle duedate */
    if (!strcmp(todoptr->duedate,"NODATE")) {
      m.indefinite = 1;
    }
    else {
      char dstr[15];

      m.indefinite = 0;
      memset(dstr, 15, 0);
      strncpy(dstr, todoptr->duedate, 2);
      m.due.tm_mday = atoi(dstr);
      memset(dstr, 15, 0);
      strncpy(dstr, todoptr->duedate + 3, 2); /* forward DD/ */
      m.due.tm_mon = atoi(dstr) - 1;   /* because months are 0..11 */
      memset(dstr, 15, 0);
      strncpy(dstr, todoptr->duedate + 6, 4); /* forward DD/MM/ */
      m.due.tm_year = atoi(dstr) - 1900;
      m.due.tm_hour = m.due.tm_min = m.due.tm_sec = 0;
      m.due.tm_isdst = -1;

      mktime(&m.due);
    }

/* handle int vars (prio, compl)*/
    m.complete = todoptr->check;
    m.priority = todoptr->priority;

/* bytepack */
    len = pack_ToDo(&m, buffer, 0xffff);

/* data is ready to send to pilot */
    dlp_WriteRecord(sd, db, todoptr->attr & dlpRecAttrSecret, todoptr->id,
      todoptr->category, buffer, len, &ret_id);
    if (todoptr->id == 0) { /* pilot returns new record's id */
      aux->id = ret_id;
      aux=aux->next;
    }

    todoptr = todoptr->next;
  }

/* write local data back to PC */
  write_todos();

  sprintf(tcl_string, "progress -2 %d \"Synchronizing\"", 200);
  code = Tcl_Eval(g_interp, tcl_string);

  if (synchType != PC2PILOT && synchType != DO_NOTHING)
    dlp_ResetSyncFlags(sd, db);

/* Close the database */
  dlp_CloseDB(sd, db);

  strcat(logbuffer, "OK ToDo"); 
  switch (synchType) {
    case PC2PILOT:   strcat(logbuffer," Pilot<<-Desktop"); break;
    case PILOT2PC:   strcat(logbuffer," Pilot->>Desktop"); break;
    case BIDIREC:    strcat(logbuffer," Pilot<->Desktop"); break; 
  }
  strcat(logbuffer, "\n");

  js_ToDo_list_free(&ToDoList_l);
  js_ToDo_list_free(&ToDoList_p);

  return 0;
}
