#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "date-lists.h"
#include <pi-dlp.h>

void js_Datebook_list_insert(struct Datebook_List **l, struct Datebook_List *elem)
{
  int i;

  if ((*l == NULL) || (*l && ((*l)->id >= elem->id))) {
    struct Datebook_List *v = (struct Datebook_List *)malloc(sizeof(struct Datebook_List));

    v->id = elem->id;
    v->category = elem->category;
    v->attr = elem->attr;

    if (elem->i_id)
      v->i_id = strdup(elem->i_id);
    else {
      v->i_id = (char *)malloc(1);
      v->i_id[0] = 0;
    }
    v->event = elem->event;
    v->begin = elem->begin;
    v->end = elem->end;
    v->alarm = elem->alarm;
    v->advance = elem->advance;
    v->advanceUnits = elem->advanceUnits;
    v->repeatType = elem->repeatType;
    v->repeatForever = elem->repeatForever;
    v->repeatEnd = elem->repeatEnd;
    v->repeatFrequency = elem->repeatFrequency;
    v->repeatDay = elem->repeatDay;

    for (i = 0; i < 7; i++)
      v->repeatDays[i] = elem->repeatDays[i];

    v->repeatWeekstart = elem->repeatWeekstart;

    v->todo = 0;
    v->done = 0;
    v->remind = elem->remind;
    v->hilite = elem->hilite;

    v->exceptions = elem->exceptions;
    if (v->exceptions)
      v->exception = (time_t *)malloc(sizeof(time_t) * elem->exceptions);
    else
      v->exception = NULL;
    for (i = 0; i < elem->exceptions; i++)
      v->exception[i] = elem->exception[i];

    if (elem->description)
      v->description = strdup(elem->description);
    else {
      v->description = (char *)malloc(1);
      v->description[0] = 0;
    }

    if (elem->note)
      v->note = strdup(elem->note);
    else {
      v->note = (char *)malloc(1);
      v->note[0] = 0;
    }

    v->next = *l;
    *l = v;
  }
  else js_Datebook_list_insert(&(*l)->next, elem);
}

void js_Datebook_list_modify(struct Datebook_List *dest, struct Datebook_List *from)
{
  int i;

  free(dest->description);
  free(dest->note);
  if (dest->exception) free(dest->exception);

  dest->id = from->id;
  dest->category = from->category;
  dest->attr = from->attr & (~dlpRecAttrDirty);

  dest->event = from->event;
  dest->begin = from->begin;
  dest->end = from->end;
  dest->alarm = from->alarm;
  dest->advance = from->advance;
  dest->advanceUnits = from->advanceUnits;
  dest->repeatType = from->repeatType;
  dest->repeatForever = from->repeatForever;
  dest->repeatEnd = from->repeatEnd;
  dest->repeatFrequency = from->repeatFrequency;
  dest->repeatDay = from->repeatDay;

  dest->remind = from->remind;
  dest->hilite = from->hilite;

  for (i = 0; i < 7; i++)
    dest->repeatDays[i] = from->repeatDays[i];

  dest->repeatWeekstart = from->repeatWeekstart;

  dest->exceptions = from->exceptions;
  if (dest->exceptions)
    dest->exception = (time_t *)malloc(sizeof(time_t) * from->exceptions);
  else
    dest->exception = NULL;
  for (i = 0; i < from->exceptions; i++)
    dest->exception[i] = from->exception[i];

  if (from->description)
    dest->description = strdup(from->description);
  else {
    dest->description = (char *)malloc(1);
    dest->description[0] = 0;
  }

  if (from->i_id)
    dest->i_id = strdup(from->i_id);
  else {
    dest->i_id = (char *)malloc(1);
    dest->i_id[0] = 0;
  }

  if (from->note)
    dest->note = strdup(from->note);
  else {
    dest->note = (char *)malloc(1);
    dest->note[0] = 0;
  }
}

int js_Datebook_list_compare(struct Datebook_List *l, struct Datebook_List *l1)
{
  int i;

  if (!(
l->event == l1->event &&
l->begin == l1->begin &&
l->end == l1->end &&
l->alarm == l1->alarm &&
l->advance == l1->advance &&
l->advanceUnits == l1->advanceUnits &&
l->repeatType == l1->repeatType &&
l->repeatForever == l1->repeatForever &&
l->repeatEnd == l1->repeatEnd &&
l->repeatFrequency == l1->repeatFrequency &&
l->repeatDay == l1->repeatDay &&
l->repeatWeekstart == l1->repeatWeekstart &&
l->exceptions == l1->exceptions))
  return 1;

  for (i = 0; i < 7; i++)
    if (l->repeatDays[i] != l1->repeatDays[i]) return 1;

  for (i = 0; i < l->exceptions; i++)
    if (l->exception[i] != l1->exception[i]) return 1;

  return strcmp(l->description, l1->description);
}

struct Datebook_List *js_Datebook_list_lookup(struct Datebook_List *l, recordid_t id)
{
  if (l == NULL) return NULL;
  else if (l->id == 0) return NULL;
  else if (l->id == id) return l;
  else return js_Datebook_list_lookup(l->next, id);
}

int js_Datebook_list_count(struct Datebook_List *l)
{
  if (l == NULL) return 0;
  else return 1 + js_Datebook_list_count(l->next);
}

void js_Datebook_list_remove(struct Datebook_List **l, struct Datebook_List *elem)
{
  if (*l == NULL) return;
  else if (elem->id == (*l)->id) {
    struct Datebook_List *v = *l;

    *l = (*l)->next;
    v->next = NULL;
    free(v->i_id);
    free(v->description);
    free(v->note);
    if (v->exception)
      free(v->exception);
    free(v);
  }
  else js_Datebook_list_remove(&(*l)->next, elem);
}

void js_Datebook_list_free(struct Datebook_List **l)
{
  if (*l == NULL) return;
  else {
    js_Datebook_list_free(&(*l)->next);
    free((*l)->i_id);
    free((*l)->description);
    free((*l)->note);
    if ((*l)->exception)
      free((*l)->exception);
    free(*l);
  }
}


/* formats a struct tm into a string */
void time_fmt(char *str, int siz, struct tm *t)
{
  memset(str, 0, siz);
  strftime(str, siz, "%a %b %d %H:%M:%S %Y", t);
}


/* shows the internal representation of a Datebook entry*/
void SHOW_Datebook_LIST(struct Datebook_List *l)
{
  struct Datebook_List *temp;
  char time_str[31];
  char c;

  temp=l;
  while (temp!=NULL) {
    time_fmt(time_str, 30, localtime(&temp->begin));

    if (temp->attr & dlpRecAttrDeleted) c = 'D';
    else if (temp->attr & dlpRecAttrDirty) c = 'Y';
    else c = '-';

    printf("\nID: %ld  event=%d  attr=%c\nBegin=%s\n",
           temp->id, temp->event, c, time_str);

    time_fmt(time_str, 30, localtime(&temp->end));
    printf("End=%s\n", time_str);

    time_fmt(time_str, 30, localtime(&temp->repeatEnd));
    printf("Contents: %s\nRepeat: type:%d [end=%s forever:%d freq:%d day:%d\n",
           temp->description,
           temp->repeatType,
           time_str,
           temp->repeatForever, temp->repeatFrequency,
           temp->repeatDay);
    printf("Alarm: [[%d %d]]\n",temp->advance, temp->advanceUnits);

    if (temp->exceptions >= 1) {
      time_fmt(time_str, 30, localtime(&temp->exception[0]));
      printf("Exceptions1: %s\n", time_str);
    }
    else
      printf("Exceptions1: NONE\n");
    if (temp->exceptions >= 2) {
      time_fmt(time_str, 30, localtime(&temp->exception[1]));
      printf("Exceptions2: %s\n", time_str);
    }
    else
      printf("Exceptions2: NONE\n");
    if (temp->exceptions >= 3) {
      time_fmt(time_str, 30, localtime(&temp->exception[2]));
      printf("Exceptions3: %s\n", time_str);
    }
    else
      printf("Exceptions3: NONE\n");

    printf("\n");
    temp=temp->next;
  }
  printf("//\n");
}
