#include "ical.h"
#include "Random.h"
#include "assert.h"

extern char home_dir[256];
extern char username[256];

char *d_type_str[] = {"Empty", "Single", "Days", "Months",
                      "WeekDays ", "MonthDays", "ComplexMonths",
                      "Start", "Finish", "Deleted"};

#define HOSTNAME_MAXLEN 64
#define ICAL_MAJOR 2
#define ICAL_MINOR 0

void generate_uid(char *id)
{
  char hostname[HOSTNAME_MAXLEN];
  unsigned int rand_1, rand_2 = 0;
  int sec;

  gethostname(hostname, HOSTNAME_MAXLEN);
  strtok(hostname,".");

  /* ensure a 8-hex digit string */
  rand_1 = getrandom(~(1 << ((8 * sizeof(unsigned int)) - 1)));

  /* ensure a 4-hex digit string */
  while (rand_2 <= 0x0fff)
    rand_2 = getrandom(0xffff);

  /* ensure a 1 digit value */
  sec = getrandom(9);
  sprintf(id, "%s_%x_%x_%d", hostname, (int)rand_1, (int)rand_2, sec);
  return;
}


int parse_dates_section(struct i_cal *item)
{
  char *item_d;
  char *str;
  int i = 0;

  if (!item->dates[0]) return 0;

  item_d = strdup(item->dates[0]);
  str = strtok(item_d, " ");

  item->date_type = (int *) malloc((i + 1)*sizeof(int));

/* data_type format is:  (according to i-cal source code)
        Empty
 *   |  Single <date>
 *   |  Days <date> <int>
 *   |  Months <date> <int>   
 *   |  WeekDays <set> Months <set>
 *   |  MonthDays <set> Months <set>

    Apparently ComplexMonths can also exist (...)
*/

  if (!strcmp(str, "Single")) {
    item->date_type[i] = Single;
    item->dates[i] = strdup(strtok(NULL, " "));
    item->date_freq = 1;
    item->dates[i+1] = 0; /* last element */
    return 0;
  } else if (!strcmp(str, "Empty")) {
    item->date_type[i] = Empty;
    item->dates[i] = 0;
    item->date_freq = -1;
    return 0;
  } else if (!strcmp(str, "Days")) {
    item->date_type[i] = Days;
    item->dates[i] = strdup(strtok(NULL, " "));
    item->date_freq = atoi(strtok(NULL, " \n"));
  } else if (!strcmp(str, "Months")) {
    item->date_type[i] = Months;
    item->dates[i] = strdup(strtok(NULL, " "));
    item->date_freq = atoi(strtok(NULL, " \n"));
  } else if (!strcmp(str, "WeekDays")) {
    item->date_type[i] = WeekDays;
    item->dates[i] = strdup(strtok(NULL, "\n"));
    item->date_freq = -1;
  } else if (!strcmp(str, "MonthDays")) {
    item->date_type[i] = MonthDays;
    item->dates[i] = strdup(strtok(NULL, "\n"));
    item->date_freq = -1;
  } else if (!strcmp(str, "ComplexMonths")) {
    item->date_type[i] = ComplexMonths;
    item->dates[i] = strdup(strtok(NULL, "\n"));
    item->date_freq = -1;
  }

  while (1) {
    i++;
    str = strtok(NULL, " \n");
    if (!strncasecmp(str,"End",3)) {
      item->dates = (char **) realloc(item->dates, (i+1) *sizeof(char *));
      item->dates[i] = NULL; /* last element */
      return 0;
    }

    item->date_type = (int *) realloc(item->date_type, (i+1)*sizeof(int));
    item->dates = (char **) realloc(item->dates, (i+1)*sizeof(char *));

    if (!strcmp(str, "Start")) {
      item->date_type[i] = Start;
      item->dates[i] = strdup(strtok(NULL, " \n"));
    }
    if (!strcmp(str, "Finish")) {
      item->date_type[i] = Finish;
      item->dates[i] = strdup(strtok(NULL, " \n"));
    }
    if (!strcmp(str, "Deleted")) {
      item->date_type[i] = Deleted;
      item->dates[i] = strdup(strtok(NULL, " \n"));
    }
  }

  return 0;
}

int line_matches(char *line, char *entry)
{
  char *str = strdup(line);

  return !strcasecmp(entry, (char *)strtok(str," "));
}

char *line_content(FILE *fp, char *line)
{
  char *ptr = line;
  char *start, *end;

  while (*ptr != '[') ptr++;
  start = ptr + 1; /* mark start just after '[' */

  while (1) {
    while (*ptr != ']') {
      ptr++;
      if (*ptr == '\n') { /* line ended and still no ']', read next line */
        char str[MAXLINE];
        fgets(str,MAXLINE,fp);
        strcat(line, str);
      }
    }

/* '\]' means it is the character ']' and can also exist on the text */
    if (*(ptr-1) != '\\') break;
    ptr++;
  }
  end = ptr - 1;   /* mark end just before ']' */

  ptr = strdup(start); /* copy the substring between [] */
  ptr[end - start + 1] = '\0';   /* mark end of string */
  return ptr;
}


int dump_item(struct i_cal *item)
{
  int i;

  printf("Type is %s
Start='%s'  Length='%s'  UID='%s'
Owner='%s'  Contents='%s'
Text='%s'   Todo='%s'
Done='%s'  Remind='%s'
Alarm='%s'  Hilite='%s'
Dates:\n",
  item->type==1?"Appt":"Note",
  item->start?item->start:"",
  item->length?item->length:"",
  item->uid?item->uid:"",
  item->owner?item->owner:"",
  item->contents?item->contents:"",
  item->text?item->text:"",
  item->todo?item->todo:"",
  item->done?item->done:"",
  item->remind?item->remind:"",
  item->alarm?item->alarm:"",
  item->hilite?item->hilite:"");

  for (i=0;;i++) {
    if (!item->dates[i]) break;
    printf("\tType: %s   %s\n",
           d_type_str[item->date_type[i]], item->dates[i]);
  }

  return 0;
}




int write_ical_header(char *location, int weekstarts)
{
  FILE *fp=fopen(location,"w");

  if (!fp) return -1;
  fprintf(fp,"Calendar [v%d.%d]\n", ICAL_MAJOR, ICAL_MINOR);
  fprintf(fp,"MondayFirst [%d]\n", weekstarts ? 1 : 0);

  fclose(fp);
  return 0;
}


int write_ical_entry(char *location, struct i_cal *item)
{
  FILE *fp=fopen(location,"a");
  if (!fp) return -1;

  switch (item->type) {
    case 0:
      fprintf(fp, "Unknown [\n");
    break;
    case 1:
      fprintf(fp, "Appt [\n");
    break;
    case 2:
      fprintf(fp, "Note [\n");
    break;
  }

  if (item->start && item->type == 1)
    fprintf(fp, "Start [%s]\n", item->start);
  if (item->length)
    fprintf(fp, "Length [%s]\n", item->length);
  if (item->uid)
    fprintf(fp, "Uid [%s]\n", item->uid);
  if (item->owner)
    fprintf(fp, "Owner [%s]\n", item->owner);
  if (item->contents)
    fprintf(fp, "Contents [%s]\n", item->contents);
/*  if (item->text)
    fprintf(fp, "Text [%s]\n", item->text);*/
  if (item->todo)
    fprintf(fp, "Todo [%s]\n", item->todo);
  if (item->done)
    fprintf(fp, "Done [%s]\n", item->done);
  if (item->remind)
    fprintf(fp, "Remind [%s]\n", item->remind);
  if (item->alarm && item->type == 1)
    fprintf(fp, "Alarms [%s]\n", item->alarm);
  if (item->hilite)
    fprintf(fp, "Hilite [%s]\n", item->hilite);
  if (item->dates[0]) {
    int i;

  /* write date header */
    fprintf(fp, "Dates [");

  /* write date type (Single, Days...) */
    fprintf(fp, "%s %s",
            d_type_str[item->date_type[0]], item->dates[0]);

  /* types = {Days, Months} must include frequency indicator */
    if (item->date_type[0] == Days ||
        item->date_type[0] == Months)
      fprintf(fp, " %d", item->date_freq);

  /* handle all other date entries if they exist (Start, Finish, Deleted) */
    for (i=1;;i++) {
      if (!item->dates[i]) { /* we've reached the end, mark it on file */
        fprintf(fp, " End\n");
        break;
      }
      /* write this date spec to file */
      fprintf(fp, "\n%s %s",
           d_type_str[item->date_type[i]], item->dates[i]);
    }
  } /* if has dates */

/* finish off date section */
  fprintf(fp, "]\n]\n");
  fclose(fp);

  return 0;
}


/* reads the calendar file's header -- catch mondayfirst */
int read_ical_header(int *mondayfirst)
{
  FILE *fp;
  char line[MAXLINE], location[256];

  sprintf(location, "%s/%s/conf/jetdate.dat", home_dir, username);
  fp = fopen(location,"r");
  WARN(fp)
  fgets(location, 80, fp); /* read location of i.cal file */
  fclose(fp);
  location[strlen(location)-1]=0;

  strcat(location,"/i.cal");
  fp = fopen(location,"r");
  WARN(fp)
  if (!fp || feof(fp)) return 0; /* file is empty */

  fgets(line,MAXLINE,fp); /* Discard first line ("Calendar [vX.Y]") */

  while (1) {
    memset(line, 0, MAXLINE);
    fgets(line,MAXLINE,fp);

    if (feof(fp)) break;
    if (line_matches(line,"Appt")) break;
    if (line_matches(line,"Note")) break;

    if (line_matches(line,"MondayFirst")) {
      *mondayfirst = atoi(line_content(fp, line));
      return 0;
    }
  }
  *mondayfirst = 0;
  fclose(fp);
  return 0;
}


/* reads the next entry in the iCal file (assumes fp points to a open file) */
int read_ical_entry(FILE *fp, struct i_cal *item)
{
  char line[MAXLINE];

  item->start = 
      item->length = 
      item->uid = 
      item->owner = 
      item->contents = 
      item->text = 
      item->todo = 
      item->done = 
      item->remind = 
      item->alarm = 
      item->hilite = 0;
    item->dates = NULL;
    item->date_type = NULL;

  while (!feof(fp)) {
    memset(line, 0, MAXLINE);
    fgets(line,MAXLINE,fp);
    if (feof(fp)) break;

    if (line_matches(line,"Appt"))
      item->type = 1;
    else if (line_matches(line,"Note"))
      item->type = 2;

    else if (line_matches(line,"Start"))
      item->start = strdup(line_content(fp, line));
    else if (line_matches(line,"Length"))
      item->length = strdup(line_content(fp, line));
    else if (line_matches(line,"Uid"))
      item->uid = strdup(line_content(fp, line));
    else if (line_matches(line,"Owner"))
      item->owner = strdup(line_content(fp, line));
    else if (line_matches(line,"Contents"))
      item->contents = strdup(line_content(fp, line));
    else if (line_matches(line,"Text"))
      item->text = strdup(line_content(fp, line));
    else if (line_matches(line,"Todo"))
      item->todo = strdup(line_content(fp, line));
    else if (line_matches(line,"Done"))
      item->done = strdup(line_content(fp, line));
    else if (line_matches(line,"Remind"))
      item->remind = strdup(line_content(fp, line));
    else if (line_matches(line,"Alarms"))
      item->alarm = strdup(line_content(fp, line));
    else if (line_matches(line,"Hilite"))
      item->hilite = strdup(line_content(fp, line));
    else if (line_matches(line,"Dates")) {
      item->dates = (char **) malloc(sizeof(char *));
      item->dates[0] = strdup(line_content(fp, line));
    }

    else if (line_matches(line,"]\n")) {
      parse_dates_section(item);
      return 0;
    }
  }

  return -1;
}
