#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mail-lists.h"
#include <pi-dlp.h>

void js_Mail_list_insert(struct Mail_List **l, struct Mail_List *elem)
{
  if (!*l || (*l && ((*l)->date >= elem->date))) {
    struct Mail_List *v = (struct Mail_List *)malloc(sizeof(struct Mail_List));

    v->id = elem->id;
    v->popid = elem->popid;
    v->category = elem->category;
    v->attr = elem->attr;

    v->dated = elem->dated;
    v->date = elem->date;

    if (elem->subject)
      v->subject = strdup(elem->subject);
    else {
      v->subject = (char *)malloc(1);
      v->subject[0] = 0;
    }

    if (elem->from)
      v->from = strdup(elem->from);
    else {
      v->from = (char *)malloc(1);
      v->from[0] = 0;
    }

    if (elem->to)
      v->to = strdup(elem->to);
    else {
      v->to = (char *)malloc(1);
      v->to[0] = 0;
    }

    if (elem->cc)
      v->cc = strdup(elem->cc);
    else {
      v->cc = (char *)malloc(1);
      v->cc[0] = 0;
    }

    if (elem->bcc)
      v->bcc = strdup(elem->bcc);
    else {
      v->bcc = (char *)malloc(1);
      v->bcc[0] = 0;
    }

    if (elem->replyTo)
      v->replyTo = strdup(elem->replyTo);
    else {
      v->replyTo = (char *)malloc(1);
      v->replyTo[0] = 0;
    }

    if (elem->sentTo)
      v->sentTo = strdup(elem->sentTo);
    else {
      v->sentTo = (char *)malloc(1);
      v->sentTo[0] = 0;
    }

    if (elem->body)
      v->body = strdup(elem->body);
    else {
      v->body = (char *)malloc(1);
      v->body[0] = 0;
    }

    v->next = *l;
    *l = v;
  }
  else js_Mail_list_insert(&(*l)->next, elem);
}

void js_Mail_list_modify(struct Mail_List *dest, struct Mail_List *from)
{
    if (dest->subject) free(dest->subject);
    if (dest->from) free(dest->from);
    if (dest->to) free(dest->to);
    if (dest->cc) free(dest->cc);
    if (dest->bcc) free(dest->bcc);
    if (dest->replyTo) free(dest->replyTo);
    if (dest->sentTo) free(dest->sentTo);
    if (dest->body) free(dest->body);

    dest->category = from->category;
    dest->attr = from->attr & (~dlpRecAttrDirty);

    dest->date = from->date;
    dest->dated = from->dated;
    dest->popid = from->popid;

    if (from->subject)
      dest->subject = strdup(from->subject);
    else {
      dest->subject = (char *)malloc(1);
      dest->subject[0] = 0;
    }

    if (from->from)
      dest->from = strdup(from->from);
    else {
      dest->from = (char *)malloc(1);
      dest->from[0] = 0;
    }

    if (from->to)
      dest->to = strdup(from->to);
    else {
      dest->to = (char *)malloc(1);
      dest->to[0] = 0;
    }

    if (from->cc)
      dest->cc = strdup(from->cc);
    else {
      dest->cc = (char *)malloc(1);
      dest->cc[0] = 0;
    }

    if (from->bcc)
      dest->bcc = strdup(from->bcc);
    else {
      dest->bcc = (char *)malloc(1);
      dest->bcc[0] = 0;
    }

    if (from->replyTo)
      dest->replyTo = strdup(from->replyTo);
    else {
      dest->replyTo = (char *)malloc(1);
      dest->replyTo[0] = 0;
    }

    if (from->sentTo)
      dest->sentTo = strdup(from->sentTo);
    else {
      dest->sentTo = (char *)malloc(1);
      dest->sentTo[0] = 0;
    }

    if (from->body)
      dest->body = strdup(from->body);
    else {
      dest->body = (char *)malloc(1);
      dest->body[0] = 0;
    }
}

int js_Mail_list_compare(struct Mail_List *l, struct Mail_List *l1)
{
  return
    strcmp(l->from, l1->from) || strcmp(l->to, l1->to) ||
    strcmp(l->subject, l1->subject) || strcmp(l->cc, l1->cc) ||
    strcmp(l->bcc, l1->bcc) || strcmp(l->replyTo, l1->replyTo) ||
    strcmp(l->sentTo, l1->sentTo) || strcmp(l->body, l1->body);
}

struct Mail_List *js_Mail_list_lookup(struct Mail_List *l, time_t date)
{
  if (l == NULL) return NULL;
  else if (l->date == date) return l;
  else return js_Mail_list_lookup(l->next, date);
}

int js_Mail_list_count(struct Mail_List *l)
{
  if (l == NULL) return 0;
  else return 1 + js_Mail_list_count(l->next);
}

void js_Mail_list_remove(struct Mail_List **l, struct Mail_List *elem)
{
  if (*l == NULL) return;
  else if (elem->date == (*l)->date && !js_Mail_list_compare(elem, *l)) {
    struct Mail_List *v = *l;
    *l = (*l)->next;
    v->next = NULL;
    js_Mail_list_free(&v);
  }
  else js_Mail_list_remove(&(*l)->next, elem);
}

void js_Mail_list_free(struct Mail_List **l)
{
  if (*l == NULL) return;
  else {
    js_Mail_list_free(&(*l)->next);
    if ((*l)->subject) free((*l)->subject);
    if ((*l)->from) free((*l)->from);
    if ((*l)->to) free((*l)->to);
    if ((*l)->cc) free((*l)->cc);
    if ((*l)->bcc) free((*l)->bcc);
    if ((*l)->replyTo) free((*l)->replyTo);
    if ((*l)->sentTo) free((*l)->sentTo);
    if ((*l)->body) free((*l)->body);

    free(*l);
  }
}


void SHOW_Mail_LIST(struct Mail_List *l)
{
  struct Mail_List *temp;

  temp=l;
  while (temp!=NULL) {
  printf("\n[cat:%d] From:%s\nDate:%ld\nSubject:%s\n",
     temp->category, temp->from,
     temp->date, temp->subject);
  temp=temp->next;
  }
  printf("//\n");
}
