#include "mail.h"

extern char username[256];

int popfd;

int getpopchar(int socket)
{
  unsigned char buf;
  int l;
  
  do {
    l = read(socket, &buf, 1);
    if (l < 0)
      return l;
  } while (!l || buf == '\r');
  
  return buf;
}


int getpopstring(int socket, char * buf)
{
  int c;

  while ((c = getpopchar(socket)) >= 0) {
    *buf++ = c;
    if (c == '\n')
      break;
  }
  *buf = '\0';
  return c;
}


char *skipspace(char * c) {
  while (c && ((*c == ' ') || (*c == '\t')))
    c++;
  return c;
}


int header(struct Mail * m, char * t)
{
  static char holding[4096];
  int skip=0;

  if (t && strlen(t) && t[strlen(t)-1] == '\n')
    t[strlen(t)-1] = 0;
  
  if (t && ((t[0] == ' ') || (t[0] == '\t'))) {
    if ((strlen(t) + strlen(holding)) > 4096)
      return 0; /* Just discard approximate overflow */
    strcat(holding, t+1);
    return 0;
  }
  
  /* Decide on what we do with m->sendTo */
  if (strncmp(holding, "Content-Type:", 13)==0) {
    if (strncasecmp("text/plain;", skipspace(holding+13), 11)) {
      skip=-1;
    }
    if (!strncasecmp("text", skipspace(holding+13), 4)) {
      skip=0;
    }
  }
  
  if (strncmp(holding, "From:", 5)==0)
    m->from = strdup(skipspace(holding+5));
  else if (strncmp(holding, "To:",3)==0)
    m->to = strdup(skipspace(holding+3));
  else if (strncmp(holding, "Subject:",8)==0)
    m->subject = strdup(skipspace(holding+8));
  else if (strncmp(holding, "Cc:",3)==0)
    m->cc = strdup(skipspace(holding+3));
  else if (strncmp(holding, "Bcc:",4)==0)
    m->bcc = strdup(skipspace(holding+4));
  else if (strncmp(holding, "Reply-To:",9)==0)
    m->replyTo = strdup(skipspace(holding+9));
  else if (strncmp(holding, "Date:",4)==0) {
    time_t d = parsedate(skipspace(holding+5));
    if (d != -1) {
      struct tm * d2; /* save date */
      d2 = localtime(&d);
      d2->tm_isdst = -1;
      m->date = *d2;

      m->dated = 1; /* signal has date */
    }
  }
  
  holding[0] = 0;
  
  if (t)
    strcpy(holding, t);

  return skip;
}


int init_pop(char *pop_host, char *pop_user, char*pop_pass)
{
  struct sockaddr_in pop_addr;
  struct hostent *hostent;
  char buffer[0xffff];

  memset((char*)&pop_addr, 0, sizeof(pop_addr));
  if ((pop_addr.sin_addr.s_addr = inet_addr(pop_host)) == -1) {
    if (!(hostent = gethostbyname(pop_host))) {
      fprintf(stderr, "Unable to resolve POP host '%s'", pop_host);
      return -1;
    }
    memcpy((char*)&pop_addr.sin_addr.s_addr, hostent->h_addr,hostent->h_length);
  }
  pop_addr.sin_family = AF_INET;
  pop_addr.sin_port = htons(110); /* POP port */
  
  if ((popfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("Unable to obtain socket");
    return -1;
  }
  
  if (connect(popfd, (struct sockaddr*)&pop_addr, sizeof(pop_addr))<0) {
    fprintf(stderr, "Unable to connect to POP server '%s'\n", pop_host);
    close(popfd);
    return -1;
  }

  read(popfd, buffer, 1024);
  if (buffer[0] != '+') {
    fprintf(stderr, "POP server failed to announce itself\n");
    close(popfd);
    return -1;
  }  
  
  sprintf(buffer, "USER %s\r\n", pop_user);
  write(popfd, buffer, strlen(buffer));
  read(popfd, buffer, 1024);
  if (buffer[0] != '+') {
    fprintf(stderr, "USER command to POP server failed\n");
    close(popfd);
    return -1;
  }

  sprintf(buffer, "PASS %s\r\n", pop_pass);
  write(popfd, buffer, strlen(buffer));
  read(popfd, buffer, 1024);
  if (buffer[0] != '+') {
    fprintf(stderr, "PASS command to POP server failed\n");
    close(popfd);
    return -1;
  }

  return 0;
}


int dele_popmail(int i)
{
  char buffer[1024];

  sprintf(buffer, "DELE %d\r\n", i);
  write(popfd, buffer, strlen(buffer));
  read(popfd, buffer, 1024);

  return 0;
}


int cmp_mail(struct Mail *s, struct Mail *t)
{
  if (t->subject) {
    if (strcmp(s->subject, t->subject)) return 1;
  }
  else if (s->subject[0]) return 1;
  if (t->from) {
    if (strcmp(s->from, t->from)) return 1;
  }
  else if (s->from[0]) return 1;
  if (t->to) {
    if (strcmp(s->to, t->to)) return 1;
  }
  else if (s->to[0]) return 1;
  if (t->cc) {
    if (strcmp(s->cc, t->cc)) return 1;
  }
  else if (s->cc[0]) return 1;
  if (t->bcc) {
    if (strcmp(s->bcc, t->bcc)) return 1;
  }
  else if (s->bcc[0]) return 1;
  if (t->replyTo) {
    if (strcmp(s->replyTo, t->replyTo)) return 1;
  }
  else if (s->replyTo[0]) return 1;
  if (t->sentTo) {
    if (strcmp(s->sentTo, t->sentTo)) return 1;
  }
  else if (s->sentTo[0]) return 1;
  if (t->body) {
    if (strcmp(s->body, t->body)) return 1;
  }
  else if (s->body[0]) return 1;

  return 0;
}

int read_popmail(int i, struct Mail *t)
{
  char buffer[0xffff];
  int l;
  int h, cnt;
  int len;
  char * msg;

  t->to = 0;
  t->from = 0;
  t->cc = 0;
  t->bcc = 0;
  t->subject = 0;
  t->replyTo = 0;
  t->sentTo = 0;
  t->body = 0;
  t->dated = 0;
    
  sprintf(buffer, "LIST %d\r\n", i);
  write(popfd, buffer, strlen(buffer));
  l = read(popfd, buffer, 1024);
  if (l < 0) {
    printf("Unable to read from socket\n");
    return -1;
  }
  buffer[l] = 0;
  if (buffer[0] != '+')
    return -2; /* signal for break */
    
  sscanf(buffer, "%*s %*d %d", &len);
    
  sprintf(buffer, "RETR %d\r\n", i);
  write(popfd, buffer, strlen(buffer));
  l = getpopstring(popfd, buffer);
  if ((l < 0) || (buffer[0] != '+')) {
    /* Weird */
    return -3; /* signal for continue */
  }
  else
    buffer[l] = 0;
    
  msg = (char *)buffer;
  h = 1;
  cnt = l;
  for(;;) {
    if (h==2) {
      /* Message was not text/plain, "fast-forward" it to its end  */
      /* This is used so that we can keep popfd synced */
      while (1) {
        memset(buffer, 0, 8192);
        cnt=read(popfd, buffer, 8192);

        if ((buffer[cnt-3] == '.')
          && (buffer[cnt-2] == '\n' || buffer[cnt-2] == '\r')
          && (buffer[cnt-1] == '\n' || buffer[cnt-1] == '\r')
          && (buffer[cnt] == '\0'))
          break;
      } /* end while */

      /* popfd now points to the beginning of a new message */
      break;
    } /* end if */


    /* read one line from the popfd */
    if ((cnt += getpopstring(popfd, msg)) < 0) {
      fprintf(stderr, "Error reading message\n");
      return -1;
    }
      
    if (h == 1) { /* Header mode */
      if ((msg[0] == '.') && (msg[1] == '\n') && (msg[2] == 0))
        break; /* end of mail */

      if (msg[0] == '\n') /* a simple return means end of headers */
        h = 0;
      else if (header(t, msg)==-1) /* test if it is a header */
        h = 2;                /* header Content-type is not text/plain */
      continue;
    }
      
    if ((msg[0] == '.') && (msg[1] == '\n') && (msg[2] == 0)) {
      msg[0] = 0;
      break; /* end of mail */
    }
      
    if (msg[0] == '.') {
      /* Must be escape */
      memmove(msg, msg+1, strlen(msg));
    }
      
    msg += strlen(msg);
  } /* for */

  /* finished retrieving an email */

  if (h == 2) /* skip if non text/plain message */
    return -3; /* signal for continue */
  else if (h) {
    /* Incomplete message, still reading headers */
    fprintf(stderr, "Incomplete message %d\n", i);
    return -3; /* signal for continue */
  }

  /* mail message is valid */

  t->body = strdup(buffer);

  return 0;
}


/* close pop */
int close_pop()
{
  char buffer[1024];

  sprintf(buffer, "QUIT\r\n");
  write(popfd, buffer, strlen(buffer));
  read(popfd, buffer, 1024);
  if (buffer[0] != '+') {
    fprintf(stderr, "QUIT command to POP server failed\n");
    return -1;
  }
  close(popfd);
  return 0;
}


/* send an email via sendmail program */
int send_mail(char *user, char *pop_host, struct Mail *t)
{
  FILE * sendf;
  char *str = NULL;

  sendf = popen(SENDMAIL, "w");

  if (!sendf) {
    fprintf(stderr,"Error launching '%s' to transmit mail!\n",
            SENDMAIL);
    return -1;
  }

  if (t->from && t->from[0])
    fprintf(sendf,"From: %s\n", t->from);
  else
    fprintf(sendf,"From: \"%s - %s\" <%s@%s>\n",
                 username, PILOTFROM, user, pop_host);

  if (t->to && t->to[0]) {
    if (!strchr(t->to,'@')) {
      str = strdup(t->to);
      sprintf(t->to,"\"%s\" <%s@%s>", str, user, pop_host);
    }
    fprintf(sendf,"To: %s\n", t->to);
  }

  if (t->dated) {
    char time_str[31];

    strftime(time_str, 30, "%a %b %d %H:%M:%S %Y", &t->date);
    fprintf(sendf,"Date: %s", time_str);
  }
  else {
    time_t tt = time(0);

    fprintf(sendf,"Date: %s", ctime(&tt));
  }

  t->dated = 1; /* signal has date */

  if (t->cc && t->cc[0])
    fprintf(sendf,"Cc: %s\n", t->cc);
  if (t->bcc && t->bcc[0])
    fprintf(sendf,"Bcc: %s\n", t->bcc);
  if (t->replyTo && t->replyTo[0])
    fprintf(sendf,"Reply-To: %s\n", t->replyTo);
  if (t->subject && t->subject[0])
    fprintf(sendf,"Subject: %s\n", t->subject);
  fprintf(sendf, "X-mailer: jetsync; version 1.0; via mail conduit\n");
  fprintf(sendf, "Content-Type: text/plain\n");
  fprintf(sendf,"\n"); /* Separate header */

  if (t->body && t->body[0]) {
    fputs(t->body, sendf);
    fprintf(sendf,"\n");
  }

  if (pclose(sendf))
    printf("Error closing %s\n", SENDMAIL);

  return 0;
}
