#include "map.h"

#define WARN(x)


/* inserts a map entry at the end of the map list */
void map_insert(map_t **m, char *i_id, recordid_t p_id)
{
/* The following if does not allow the insertion of an item that
   already exists in the map */
  if (*m && strcmp(i_id, (*m)->i_id) && (p_id != (*m)->p_id))
    map_insert(&(*m)->next, i_id, p_id);
  else {
    *m = (map_t *) malloc(sizeof(map_t));
    (*m)->i_id = strdup(i_id);
    (*m)->deleted = 1;
    (*m)->p_id = p_id;
    (*m)->next = NULL;
  }
}


/* removes a map entry from the map list */
void map_remove(map_t **m, char *i_id, recordid_t p_id)
{
  if (*m) {
    if (!strcmp((*m)->i_id, i_id) || (*m)->p_id == p_id) {
      map_t *deleted = *m;

      *m = (*m)->next;
      deleted->next = NULL;
      map_free(&deleted);
    }
    else
      map_remove(&(*m)->next, i_id, p_id);
  }
}


/* sets/unsets the deleted flag */
void map_set_del(map_t **m, char *i_id, int del)
{
  if (*m) {
    if (!strcmp((*m)->i_id, i_id))
      (*m)->deleted = del;
    else
      map_set_del(&(*m)->next, i_id, del);
  }
}

/* frees all memory allocated by a map */
void map_free(map_t **m)
{
  if (!*m) return;
  else {
    map_free(&(*m)->next);

    free((*m)->i_id);
    free(*m);
    *m = NULL;
  }
}


int map_display(map_t *map)
{
  map_t *ptr = map;

  printf("Map:\n");
  while (ptr) {
    printf("%s <-> %ld\n", ptr->i_id, ptr->p_id);
    ptr = ptr->next;
  }
  return 0;
}


/* reads the map file from disk */
int map_read(map_t **map)
{
  char location[256];
  FILE *fp;
  char *i_id;
  recordid_t p_id;
  char str[80];

  sprintf(location,"%s/%s/conf/ical.map", home_dir, username);
  fp = fopen(location,"r");

  if (!fp) return -1;
  while (1) {
    memset(str, 0, 80);
    fgets(str, 80, fp);
    if (feof(fp)) break;

    /* get ical record ID */
    i_id = strtok(str, ":");

    /* detect errors */
    if (!i_id || !i_id[1]) continue;
 
    /* get pilot record ID */
    p_id = (recordid_t) atol(strtok(NULL, ":"));

    /* valid map entry, insert it */
    map_insert(map, i_id, p_id);
  }
  fclose(fp);
  return 0;
}


/* writes the map file to disk */
int map_write(map_t *map)
{
  char location[256];
  FILE *fp;
  map_t *ptr = map;

  sprintf(location,"%s/%s/conf/ical.map", home_dir, username);
  fp = fopen(location,"w");

  if (!fp) return -1;
  while (ptr) {
    fprintf(fp,"%s:%ld\n", ptr->i_id, ptr->p_id);
    ptr = ptr->next;
  }
  fclose(fp);
  return 0;
}


/* looks for an item via its p_id and returns the correspondent i_id */
char *map_lookup_p(map_t *map, recordid_t p_id)
{
  map_t *ptr = map;

  while (ptr) {
    if (ptr->p_id == p_id) return ptr->i_id; /* ID was found */
    ptr = ptr->next;
  }
  return ""; /* ID was not found */
}


/* looks for an item via its i_id */
recordid_t map_lookup_i(map_t *map, char *i_id)
{
  map_t *ptr = map;

  while (ptr) {
    if (!strcmp(ptr->i_id, i_id)) return ptr->p_id; /* ID was found */
    ptr = ptr->next;
  }
  return -1; /* ID was not found */
}
