#include <tk.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>

#include "lists.h"
#include <pi-dlp.h>

Tk_Window mainw;
Tcl_Interp *g_interp;

struct List *AddrList=NULL;
struct List *addrptr;
int new_flag = 0;

#define MAX_BASIC_CATEGORIES 16

char basic_categories[16][16];

char datadir[80]=""; /* directory where data resides */
char home_dir[256]="";
char username[256]="";

char custom[4][16]; /* 4 custom entries */
int customflags[4];      /* 4 custom renaming flags */

int get_category(char *str)
{
  int i;

  for (i = 0; i < MAX_BASIC_CATEGORIES; i++)
    if (!strcmp(str, basic_categories[i])) return i;
  return atoi(str);
}

int write_addrs()
{
  int i;
  char location[512];
  FILE *fp;

  sprintf(location,"%s/addrs",datadir);

  if (!(fp = fopen(location,"w"))) {
    printf("Couldn't open %s for writing...\n",location);
    return -1; 
  }

  addrptr = AddrList;

/* write super-header (custom) */
  fprintf(fp,"%s\n%s\n%s\n%s\n%d\n%d\n%d\n%d\n",
             custom[0], custom[1], custom[2], custom[3],
             customflags[0], customflags[1], customflags[2], customflags[3]);

  while (addrptr != NULL) {
    /* write record */
    fprintf(fp,"%ld %d %d %d",
            addrptr->id, addrptr->category, addrptr->attr,
            addrptr->showPhone);

    for (i=0; i<5; i++)
      fprintf(fp," %d", addrptr->phoneLabel[i]);

    fprintf(fp,"\n");

    for (i=0; i<19; i++)
      fprintf(fp," %d", addrptr->size[i]);

    fprintf(fp,"\n");

    for (i=0; i<19; i++)
      fprintf(fp,"%s\n", addrptr->size[i]?addrptr->entry[i]:"");

    addrptr = addrptr->next;
  }
  fclose(fp);

  return 0;
}

void __quit()
{
  write_addrs();
  exit(0);
}

void *ctrlc(int i)
{
  __quit();
  return 0;
}

int quit(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  __quit();
  return 0;
}

int reset_addrptr(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  addrptr=AddrList;
  return 0;
}

int set_addrs(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  int i;

  if (!strcmp(argv[1],"id")) {
    if (!addrptr) {
      struct List elem;

      elem.id = 0;
      for (i=0; i<19; i++)
        elem.size[i] = 0;
      for (i=0; i<5; i++)
        elem.phoneLabel[i] = 0;
      js_list_insert(&AddrList, &elem);
      addrptr=AddrList;
      addrptr->attr = 0; /* ensure initial null attributes */
      new_flag = 1;
      return 0;
     }
   addrptr->id = atol(argv[2]);
  }
  else if (!strcmp(argv[1],"category")) {
    addrptr->category = get_category(argv[2]);
  }
  else if (!strcmp(argv[1],"private")) {
    addrptr->attr &= ~dlpRecAttrSecret;
    addrptr->attr |= (atoi(argv[2]) * dlpRecAttrSecret);
  }
  else if (!strcmp(argv[1],"deleted")) {
    addrptr->attr |= (atoi(argv[2]) * dlpRecAttrDeleted);
  }
  else if (!strcmp(argv[1],"dirty")) {
    addrptr->attr |= (atoi(argv[2]) * dlpRecAttrDirty);
  }
  else if (!strcmp(argv[1],"showphone")) {
    addrptr->showPhone = atoi(argv[2]);
  }
  else if (!strncmp(argv[1], "phonelabel", 10)) {
    i = atoi(argv[1]+10); /* get phone label index 0..4 */
    addrptr->phoneLabel[i] = atoi(argv[2]);
  }
  else if (!strncmp(argv[1], "custom", 6)) {
    i = atoi(argv[1]+6); /* get custom field names 0..4 */

    if (strcmp(custom[i], argv[2])) {
      strcpy(custom[i], argv[2]);
      customflags[i] = 1;
    }
  }
  else if (!strncmp(argv[1], "entry", 5)) {
    i = atoi(argv[1]+5); /* get entry index 0..18 */
    addrptr->size[i] = strlen(argv[2]);
    if (addrptr->size[i]) addrptr->entry[i] = strdup(argv[2]);

    if (i==18) /* last entry */
      addrptr=addrptr->next;
  }
  else
    return -1;

  return 0;
}

int get_addrs(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  interp->result = (char *)malloc(4096);
  if (!strcmp(argv[1],"id")) {
    sprintf(interp->result,"%ld",addrptr->id);
  }
  else if (!strcmp(argv[1],"category")) {
    sprintf(interp->result,"%s", addrptr->category < MAX_BASIC_CATEGORIES ?
                                 basic_categories[addrptr->category] :
                                 "-");
  }
  else if (!strcmp(argv[1],"private")) {
    sprintf(interp->result,"%d",addrptr->attr & dlpRecAttrSecret?1:0);
  }
  else if (!strcmp(argv[1],"deleted")) {
    sprintf(interp->result,"%d",addrptr->attr & dlpRecAttrDeleted?1:0);
  }
  else if (!strcmp(argv[1],"dirty")) {
    sprintf(interp->result,"%d",addrptr->attr & dlpRecAttrDirty?1:0);
  }
  else if (!strcmp(argv[1],"showphone")) {
    sprintf(interp->result,"%d",addrptr->showPhone);
  }
  else if (!strncmp(argv[1],"phonelabel", 10)) {
    int i = atoi(argv[1]+10); /* get phone label index */
    sprintf(interp->result, "%d", addrptr->phoneLabel[i]);
  }
  else if (!strncmp(argv[1],"custom", 6)) {
    int i = atoi(argv[1]+6); /* get custom field names 0..4 */

    strcpy(interp->result, custom[i]);
  }
  else if (!strncmp(argv[1],"entry", 5)) {
    int i = atoi(argv[1]+5); /* get entry index */
    if (addrptr->size[i])
      strcpy(interp->result, addrptr->entry[i]);
    else
      strcpy(interp->result, "");

    if (i==18) /* last entry */
      addrptr=addrptr->next;
  }
  else
    return -1;

  return 0;
}

int read_addrs(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  char location[512];
  FILE *fp;
  char str[1024];
  int count=0, i;
  struct List elem;

  sprintf(location,"%s/%s/conf/jetaddr.cat",home_dir,username);
  fp=fopen(location,"r");
  if (fp) {
    for (i = 0; i < MAX_BASIC_CATEGORIES; i++) {
      if (feof(fp)) break;
      fgets(basic_categories[i], 15, fp);
      basic_categories[i][strlen(basic_categories[i])-1] = 0; /* \n */
    }
    fclose(fp);
  }

  sprintf(location,"%s/addrs",datadir);
  if (!(fp = fopen(location,"r"))) {
    if (!(fp = fopen(location,"w"))) {
      printf("Couldn't create file %s...\n",location);   
      return -1;
    }
    fprintf(fp,"Custom 1\nCustom 2\nCustom 3\nCustom 4\n0\n0\n0\n0\n");
    fclose(fp);
    return 0;
  }

/* read super-header (custom) */
  fgets(custom[0],16,fp);
  custom[0][strlen(custom[0])-1] = 0;
  fgets(custom[1],16,fp);
  custom[1][strlen(custom[1])-1] = 0;
  fgets(custom[2],16,fp);
  custom[2][strlen(custom[2])-1] = 0;
  fgets(custom[3],16,fp);
  custom[3][strlen(custom[3])-1] = 0;

  fgets(str,4,fp);
  customflags[0] = atoi(str);
  fgets(str,4,fp);
  customflags[1] = atoi(str);
  fgets(str,4,fp);
  customflags[2] = atoi(str);
  fgets(str,4,fp);
  customflags[3] = atoi(str);

  while (1) {
    fgets(str,1024,fp); /* read header */

    if (feof(fp)) break;
/* read addr header*/
    sscanf(str,"%ld %d %d %d %d %d %d %d %d\n",
           &elem.id, &elem.category,&elem.attr, &elem.showPhone,
           &elem.phoneLabel[0],
           &elem.phoneLabel[1],
           &elem.phoneLabel[2],
           &elem.phoneLabel[3],
           &elem.phoneLabel[4]);

/* read entry sizes */
    fgets(str,1024,fp); /* read header */
    sscanf(str," %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
          &elem.size[0], &elem.size[1], &elem.size[2], &elem.size[3],
          &elem.size[4], &elem.size[5], &elem.size[6], &elem.size[7],
          &elem.size[8], &elem.size[9], &elem.size[10], &elem.size[11],
          &elem.size[12], &elem.size[13], &elem.size[14], &elem.size[15],
          &elem.size[16], &elem.size[17], &elem.size[18]);

/* read entries */
    for (i=0; i<19; i++) {
      elem.entry[i] = (char *)malloc(elem.size[i]+1);

      fread(elem.entry[i],elem.size[i]+1,1,fp); /* read addr entry */
      elem.entry[i][elem.size[i]]=0;
    }

/* add addr to list */
    js_list_insert(&AddrList, &elem);
    count++;
  }
/* reset addrptr */
  addrptr=AddrList;
  fclose(fp);

  sprintf(interp->result,"%d",count);
  return 0;
}

void create_commands(Tcl_Interp *interp) ;

int init(ClientData clientData, Tcl_Interp *interp,
                    int argc, char **argv)
{
  char location[512];
  FILE *fp;

/* get username */
  strcpy(username,argv[1]);   
/* get home dir */
  strcpy(home_dir,argv[2]);

  sprintf(location,"%s/%s/conf/jetaddr.dat",home_dir,username);
  fp = fopen(location,"r");
  fgets(datadir,80,fp);
  datadir[strlen(datadir)-1]=0;
  fclose(fp);

  return 0;
}

int
main(int argc, char **argv)
{
    signal(SIGINT,(void *)ctrlc);
    signal(SIGTERM,(void *)ctrlc);
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;			/* Needed only to prevent compiler warning. */
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int Tcl_AppInit(Tcl_Interp *interp)
{
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }

  mainw = Tk_MainWindow(interp);
  if (mainw == NULL) {
    return TCL_ERROR;
  }

  /*
   * Call Tcl_CreateCommand for application-specific commands, if
   * they weren't already created by the init procedures called above.
   */

/* main.c */

  Tcl_CreateCommand(interp,"init",init,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"quit",quit,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"read_addrs",read_addrs,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"get_addrs",get_addrs,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"set_addrs",set_addrs,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"reset_addrptr",reset_addrptr,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);

  Tcl_SetVar(interp,"tcl_RcFileName","~/.wishrc",TCL_GLOBAL_ONLY);

  g_interp = interp;

  return TCL_OK;
}
