#include "defs.h"
#include "Handle.h"
#include "Database.h"

/* converts a long signature word to 4 ascii chars */
unsigned char *long2str(unsigned long l)
{
  unsigned char *str = (unsigned char *) malloc(5);

  sprintf(str,"%c%c%c%c",
    (char)((l & 0xFF000000) >> 24),
    (char)((l & 0x00FF0000) >> 16),
    (char)((l & 0x0000FF00) >> 8),
    (char) (l & 0x000000FF));
  return str;
}


/* converts a long signature word to 4 ascii chars */
unsigned long str2long(unsigned char *str)
{
  return (str[0] << 24) + (str[1] << 16) + (str[2] << 8) + str[3];
}


/****** JetSync API function ******/   
/* Given a socket and a DB name, return the DBInfo structure for it */
/* Arguments: inited pilot socket reference, database name */
/* Returns: via param, database info structure */                    
/**********************************/
int js_GetDBInfo(int sd, char *name, struct DBInfo *info)
{
  if (dlp_FindDBInfo(sd, 0, 0, name, 0, 0, info) < 0) {
    printf("Unable to locate database '%s'.\n", name);
    return -1;
  }
  return 0;
}


/****** JetSync API function ******/   
/* Downloads a specified file to local directory */
/* Arguments: inited pilot socket reference, database name */
/* Returns: ignored */
/**********************************/
int js_Download(int sd, char *name)
{
  struct pi_file *f;
  struct DBInfo info;
  int err;

  js_GetDBInfo(sd, name, &info);

  f = pi_file_create(name, &info);

#ifdef __DEBUG
  printf("Downloading \"%s\"...\n",name);
  printf("Creator: '%s'\n",long2str(info.creator));
  printf("Created on: %s\n",time_str(info.createDate));
  printf("Last modified: %s\n",time_str(info.modifyDate));
#endif

  if ((err=pi_file_retrieve(f, sd, 0)) < 0)
    printf("Failed, unable to download file %s (%d)\n", name, err);
  pi_file_close(f);

  return 0;
}


/****** JetSync API function ******/   
/* Uploads a specified file to pilot */
/* Arguments: inited pilot socket reference, database name */
/* Returns: ignored */
/**********************************/
int js_Upload(int sd, char *name)
{
  struct pi_file *f;
  int err;

  f = pi_file_open(name);

#ifdef __DEBUG
  printf("Uploading \"%s\"...\n",name);
#endif

  if ((err=pi_file_install(f, sd, 0)) < 0)
    printf("Failed, unable to upload file %s (%d)\n", name, err);

  pi_file_close(f);

  return 0;
}

