#include <tk.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>

#include "lists.h"
#include <pi-dlp.h>

Tk_Window mainw;
Tcl_Interp *g_interp;

struct List *MemoList=NULL;
struct List *memoptr;

#define MAX_BASIC_CATEGORIES 16

char basic_categories[16][16];

char datadir[80]; /* directory where data resides */
char home_dir[256]="";
char username[256]="";

int get_category(char *str)
{
  int i;

  for (i = 0; i < MAX_BASIC_CATEGORIES; i++)
    if (!strcmp(str, basic_categories[i])) return i;
  return atoi(str);
}

int write_memos()
{
  char location[512];
  FILE *fp;

  sprintf(location,"%s/memos",datadir);

  if (!(fp = fopen(location,"w"))) {
    printf("Couldn't open %s for writing...\n",location);
    return -1; 
  }
  memoptr = MemoList;
  while (memoptr != NULL) {
    fprintf(fp,"%ld %d %d %d\n%s\n",
            memoptr->id, memoptr->size, memoptr->category,
            memoptr->attr, memoptr->data); /* write record */

    memoptr = memoptr->next;
  }
  fclose(fp);

  return 0;
}

void __quit()
{
  write_memos();
  exit(0);
}

void *ctrlc(int i)
{
  __quit();
  return 0;
}

int quit(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  __quit();
  return 0;
}

int reset_memoptr(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  memoptr=MemoList;
  return 0;
}

int set_memos(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  if (!strcmp(argv[1],"id")) {
    if (!memoptr) {
      struct List elem;
      elem.id = 0;
      elem.size = 0;
      js_list_insert(&MemoList, &elem);
      memoptr=MemoList;
      memoptr->attr = 0; /* ensure initial null attributes */
      memoptr->size = 0;
      return 0;
     }
    memoptr->id = atol(argv[2]);
  }
  else if (!strcmp(argv[1],"category")) {
    memoptr->category = get_category(argv[2]);
  }
  else if (!strcmp(argv[1],"private")) {
    memoptr->attr &= ~dlpRecAttrSecret;
    memoptr->attr |= (atoi(argv[2]) * dlpRecAttrSecret);
  }
  else if (!strcmp(argv[1],"deleted")) {
    memoptr->attr |= (atoi(argv[2]) * dlpRecAttrDeleted);
  }
  else if (!strcmp(argv[1],"dirty")) {
    memoptr->attr |= (atoi(argv[2]) * dlpRecAttrDirty);
  }
  else if (!strcmp(argv[1],"data")) {
    memoptr->size = strlen(argv[2]);
    memoptr->data =(char *)malloc(memoptr->size + 1);
    strcpy(memoptr->data, argv[2]);
    memoptr=memoptr->next;
  }
  else
    return -1;

  return 0;
}

int get_memos(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  interp->result = (char *)malloc(4096);
  if (!strcmp(argv[1],"id")) {
    sprintf(interp->result,"%ld",memoptr->id);
  }
  else if (!strcmp(argv[1],"size")) {
    sprintf(interp->result,"%d",memoptr->size);
  }
  else if (!strcmp(argv[1],"category")) {
    sprintf(interp->result,"%s", memoptr->category < MAX_BASIC_CATEGORIES ?
                                 basic_categories[memoptr->category] :
                                 "-");
  }
  else if (!strcmp(argv[1],"private")) {
    sprintf(interp->result,"%d",memoptr->attr & dlpRecAttrSecret?1:0);
  }
  else if (!strcmp(argv[1],"deleted")) {
    sprintf(interp->result,"%d",memoptr->attr & dlpRecAttrDeleted?1:0);
  }
  else if (!strcmp(argv[1],"dirty")) {
    sprintf(interp->result,"%d",memoptr->attr & dlpRecAttrDirty?1:0);
  }
  else if (!strcmp(argv[1],"data")) {
    sprintf(interp->result,"%s",memoptr->data);
    memoptr=memoptr->next;
  }
  else
    return -1;

  return 0;
}

int read_memos(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  char location[512];
  FILE *fp;
  char str[80];
  int count=0, i;
  struct List elem;

  sprintf(location,"%s/%s/conf/jetmemo.cat",home_dir,username);
  fp=fopen(location,"r");
  if (fp) {
    for (i = 0; i < MAX_BASIC_CATEGORIES; i++) {
      if (feof(fp)) break;
      fgets(basic_categories[i], 15, fp);
      basic_categories[i][strlen(basic_categories[i])-1] = 0; /* \n */
    }
    fclose(fp);
  }

  sprintf(location,"%s/memos",datadir);
  if (!(fp = fopen(location,"r"))) {
    if (!(fp = fopen(location,"w"))) {
      printf("Couldn't create file %s...\n",location);
      return -1; 
    }
    fclose(fp);
    return 0;
  }
  while (1) {
    fgets(str,80,fp); /* read header */

    if (feof(fp)) break;

/* read memo header*/
    sscanf(str,"%ld %d %d %d",&elem.id,&elem.size,&elem.category,&elem.attr);

/* read memo body */
    elem.data = (char *)malloc(elem.size+1);
    fread(elem.data,elem.size+1,1,fp); /* read memo in record */
    if (ferror(fp)) /* read memo in record */
      printf("Error in fread (line %d, file %s)\n",__LINE__,__FILE__);
    elem.data[elem.size]=0;

/* add memo to list */
    js_list_insert(&MemoList, &elem);
    if (elem.data) free(elem.data);
    count++;
  }
/* reset memoptr */
  memoptr=MemoList;

  fclose(fp);

  sprintf(interp->result,"%d",count);
  return 0;
}

void create_commands(Tcl_Interp *interp) ;

int init(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  char location[512];
  FILE *fp;

/* get username */
  strcpy(username,argv[1]);
/* get home dir */
  strcpy(home_dir,argv[2]);

  sprintf(location,"%s/%s/conf/jetmemo.dat",home_dir,username);
  fp = fopen(location,"r");
  if (!fp)
    printf("Couldn't open %s for reading\n",location);
  fgets(datadir,80,fp);
  datadir[strlen(datadir)-1]=0;
  fclose(fp);

  return 0;
}

int
main(int argc, char **argv)
{
    signal(SIGINT,(void *)ctrlc);
    signal(SIGTERM,(void *)ctrlc);
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;			/* Needed only to prevent compiler warning. */
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int Tcl_AppInit(Tcl_Interp *interp)
{
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }

  mainw = Tk_MainWindow(interp);
  if (mainw == NULL) {
    return TCL_ERROR;
  }

  /*
   * Call Tcl_CreateCommand for application-specific commands, if
   * they weren't already created by the init procedures called above.
   */

/* main.c */

  Tcl_CreateCommand(interp,"init",init,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"quit",quit,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"read_memos",read_memos,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"get_memos",get_memos,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"set_memos",set_memos,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"reset_memoptr",reset_memoptr,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);

  Tcl_SetVar(interp,"tcl_RcFileName","~/.wishrc",TCL_GLOBAL_ONLY);

  g_interp = interp;

  return TCL_OK;
}
