#include <tk.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>

#include "lists.h"
#include <pi-dlp.h>

Tk_Window mainw;
Tcl_Interp *g_interp;

struct List *TodoList=NULL;
struct List *todoptr;
int new_flag = 0;

#define MAX_BASIC_CATEGORIES 16

char basic_categories[16][16];

char datadir[80]=""; /* directory where data resides */
char home_dir[256]="";
char username[256]="";

int get_category(char *str)
{
  int i;

  for (i = 0; i < MAX_BASIC_CATEGORIES; i++)
    if (!strcmp(str, basic_categories[i])) return i;
  return atoi(str);
}

int write_todos()
{
  char location[512], str[15];
  FILE *fp;

  sprintf(location,"%s/todos",datadir);

  if (!(fp = fopen(location,"w"))) {
    printf("Couldn't open %s for writing...\n",location);
    return -1; 
  }

  todoptr = TodoList;

  while (todoptr != NULL) {
    if (strcmp(todoptr->duedate,"NODATE")) {
      sprintf(str,"%c%c%c%c%c%c%c%c",
        todoptr->duedate[0],
        todoptr->duedate[1],

        todoptr->duedate[3],
        todoptr->duedate[4],

        todoptr->duedate[6],
        todoptr->duedate[7],
        todoptr->duedate[8],
        todoptr->duedate[9]);
      strcpy(todoptr->duedate, str);
    }

    /* write record */
    fprintf(fp,"%ld %d %d %d %d %d %d %s\n%s\n%s\n",
            todoptr->id, todoptr->check, todoptr->category,
            todoptr->attr, todoptr->priority,
            todoptr->size_n, todoptr->size_d, todoptr->duedate,
            todoptr->note, todoptr->desc);
    todoptr = todoptr->next;
  }
  fclose(fp);

  return 0;
}

void __quit()
{
  write_todos();
  exit(0);
}

void *ctrlc(int i)
{
  __quit();
  return 0;
}

int quit(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  __quit();
  return 0;
}

int reset_todoptr(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  todoptr=TodoList;
  return 0;
}

int calc_date(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  char str[15];
  time_t t_val;
  struct tm *t;

  t_val = time(NULL);
  if (!strcmp(argv[1],"TOMORROW"))
    t_val += 24 * 60 * 60;
  else if (!strcmp(argv[1],"WEEKLT"))
    t_val += 7 * 24 * 60 * 60;
  t = localtime(&t_val);

  strftime(str, 15, "%d/%m/%Y", t);
  sprintf(interp->result, "%s",str);
  return 0;
}

int set_todos(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  if (!strcmp(argv[1],"id")) {
    if (!todoptr) {
      struct List elem;

      elem.id = 0;
      elem.size_n = 0;
      elem.size_d = 0;
      strcpy(elem.duedate,"NODATE");
      js_list_insert(&TodoList, &elem);
      todoptr=TodoList;
      todoptr->attr = 0; /* ensure initial null attributes */
      new_flag = 1;
      return 0;
     }
   todoptr->id = atol(argv[2]);
  }
  else if (!strcmp(argv[1],"category")) {
    todoptr->category = get_category(argv[2]);
  }
  else if (!strcmp(argv[1],"private")) {
    todoptr->attr &= ~dlpRecAttrSecret;
    todoptr->attr |= (atoi(argv[2]) * dlpRecAttrSecret);
  }
  else if (!strcmp(argv[1],"deleted")) {
    todoptr->attr |= (atoi(argv[2]) * dlpRecAttrDeleted);
  }
  else if (!strcmp(argv[1],"dirty")) {
    todoptr->attr |= (atoi(argv[2]) * dlpRecAttrDirty);
  }
  else if (!strcmp(argv[1],"check")) {
    todoptr->check = atoi(argv[2]);
  }
  else if (!strcmp(argv[1],"priority")) {
    todoptr->priority = atoi(argv[2]);
  }
  else if (!strcmp(argv[1],"duedate")) {
    strcpy(todoptr->duedate, argv[2]);
  }
  else if (!strcmp(argv[1],"note")) {
    todoptr->size_n = strlen(argv[2]);
    todoptr->note =(char *)malloc(todoptr->size_n + 1);
    strcpy(todoptr->note, argv[2]);
  }
  else if (!strcmp(argv[1],"desc")) {
    todoptr->size_d = strlen(argv[2]);

    todoptr->desc =(char *)malloc(todoptr->size_d + 1);
    strcpy(todoptr->desc, argv[2]);

    if (!todoptr->id && new_flag) todoptr=NULL;
    else todoptr=todoptr->next;
    new_flag = 0;
  }
  else
    return -1;

  return 0;
}

int get_todos(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  interp->result = (char *)malloc(4096);
  if (!strcmp(argv[1],"id")) {
    sprintf(interp->result,"%ld",todoptr->id);
  }
  else if (!strcmp(argv[1],"category")) {
    sprintf(interp->result,"%s", todoptr->category < MAX_BASIC_CATEGORIES ?
                                 basic_categories[todoptr->category] :
                                 "-");
  }
  else if (!strcmp(argv[1],"private")) {
    sprintf(interp->result,"%d",todoptr->attr & dlpRecAttrSecret?1:0);
  }
  else if (!strcmp(argv[1],"deleted")) {
    sprintf(interp->result,"%d",todoptr->attr & dlpRecAttrDeleted?1:0);
  }
  else if (!strcmp(argv[1],"dirty")) {
    sprintf(interp->result,"%d",todoptr->attr & dlpRecAttrDirty?1:0);
  }
  else if (!strcmp(argv[1],"check")) {
    sprintf(interp->result,"%d",todoptr->check);
  }
  else if (!strcmp(argv[1],"priority")) {
    sprintf(interp->result,"%d",todoptr->priority);
  }
  else if (!strcmp(argv[1],"duedate")) {
    strcpy(interp->result, todoptr->duedate);
  }
  else if (!strcmp(argv[1],"note")) {
    strcpy(interp->result, todoptr->note);
  }
  else if (!strcmp(argv[1],"desc")) {
    strcpy(interp->result, todoptr->desc);
    todoptr=todoptr->next;
  }
  else
    return -1;

  return 0;
}

int read_todos(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char **argv)
{
  char location[512];
  FILE *fp;
  char str[1024], auxstr[20];
  int count=0,i;
  struct List elem;

  sprintf(location,"%s/%s/conf/jettodo.cat",home_dir,username);
  fp=fopen(location,"r");
  if (fp) {
    for (i = 0; i < MAX_BASIC_CATEGORIES; i++) {
      if (feof(fp)) break;
      fgets(basic_categories[i], 15, fp);
      basic_categories[i][strlen(basic_categories[i])-1] = 0; /* \n */
    }
    fclose(fp);
  }

  sprintf(location,"%s/todos",datadir);
  if (!(fp = fopen(location,"r"))) {
    if (!(fp = fopen(location,"w"))) {
      printf("Couldn't create file %s...\n",location);   
      return -1;
    }
    fclose(fp);
    return 0;
  }
  while (1) {
    fgets(str,1024,fp); /* read header */

    if (feof(fp)) break;

/* read todo header*/
    sscanf(str,"%ld %d %d %d %d %d %d %s\n",
           &elem.id,&elem.check, &elem.category,&elem.attr,
           &elem.priority, &elem.size_n, &elem.size_d, auxstr);


    if (!strcmp(auxstr,"NODATE"))
      strcpy(elem.duedate,"NODATE");
    else
      sprintf(elem.duedate,"%c%c/%c%c/%s", auxstr[0], auxstr[1],
        auxstr[2], auxstr[3], &auxstr[4]);

/* read todo body */
    elem.note = (char *)malloc(elem.size_n+1);
    elem.desc = (char *)malloc(elem.size_d+1);

    fread(elem.note,elem.size_n+1,1,fp); /* read todo note */
    elem.note[elem.size_n]=0;

    fread(elem.desc,elem.size_d+1,1,fp); /* read todo desc*/
    elem.desc[elem.size_d]=0;

/* add todo to list */
    js_list_insert(&TodoList, &elem);
    count++;
  }
/* reset todoptr */
  todoptr=TodoList;
  fclose(fp);

  sprintf(interp->result,"%d",count);
  return 0;
}

void create_commands(Tcl_Interp *interp) ;

int init(ClientData clientData, Tcl_Interp *interp,
                    int argc, char **argv)
{
  char location[512];
  FILE *fp;

/* get username */
  strcpy(username,argv[1]);   
/* get home dir */
  strcpy(home_dir,argv[2]);

  sprintf(location,"%s/%s/conf/jettodo.dat",home_dir,username);
  fp = fopen(location,"r");
  fgets(datadir,80,fp);
  datadir[strlen(datadir)-1]=0;
  fclose(fp);

  return 0;
}

int
main(int argc, char **argv)
{
    signal(SIGINT,(void *)ctrlc);
    signal(SIGTERM,(void *)ctrlc);
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;			/* Needed only to prevent compiler warning. */
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int Tcl_AppInit(Tcl_Interp *interp)
{
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }

  mainw = Tk_MainWindow(interp);
  if (mainw == NULL) {
    return TCL_ERROR;
  }

  /*
   * Call Tcl_CreateCommand for application-specific commands, if
   * they weren't already created by the init procedures called above.
   */

/* main.c */

  Tcl_CreateCommand(interp,"init",init,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"quit",quit,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"read_todos",read_todos,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"get_todos",get_todos,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"set_todos",set_todos,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"reset_todoptr",reset_todoptr,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp,"calc_date",calc_date,(ClientData *)NULL,
                    (Tcl_CmdDeleteProc *) NULL);

  Tcl_SetVar(interp,"tcl_RcFileName","~/.wishrc",TCL_GLOBAL_ONLY);

  g_interp = interp;

  return TCL_OK;
}
